# -*- coding: utf-8 -*-
from collections import OrderedDict
import datetime, os.path
import sys, traceback
from rprj.dblayer import ForeignKey, DBEntity, DBAssociation, DBEFactory, DBMgr


# Predefined Groups
GROUP_ADMIN = '-2'
GROUP_USERS = '-3'
GROUP_GUESTS = '-4'
GROUP_PROJECT = '-5'
GROUP_WEBMASTER = '-6'

dbschema = OrderedDict()
#dbschema = {}

# TODO translate all those PHP functions: start.
# TODO translate all those PHP functions: end.

class DBEDBVersion(DBEntity):
	__columns = {}
	def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
		DBEntity.__init__(self, tablename, names, values, attrs, keys )
		if len(DBEDBVersion.__columns.keys())==0:
			parentcols = self.getColumns()
			for k in parentcols.keys():
				DBEDBVersion.__columns[k] = parentcols[k]
			DBEDBVersion.__columns['version'] = ["int","not null"]
		self._columns=DBEDBVersion.__columns
	def getTableName(self):
		return "dbversion"
	def getKeys(self):
		return { 'version':'int' }
	def getFK(self):
		return [  ]
	def getOrderBy(self):
		return [ "version" ]
	def version(self):
		tmp = self.getValue('version')
		if tmp is None:
			return 0
		return tmp
dbschema['dbversion'] = DBEDBVersion

class DBEUserGroup(DBEntity):
	__columns={}
	def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
		DBEntity.__init__(self, tablename, names, values, attrs, keys )
		if len(DBEUserGroup.__columns.keys())==0:
			parentcols = self.getColumns()
			for k in parentcols.keys():
				DBEUserGroup.__columns[k] = parentcols[k]
			DBEUserGroup.__columns['user_id'] = ["uuid","not null"]
			DBEUserGroup.__columns['group_id'] = ["uuid","not null"]
		self._columns = DBEUserGroup.__columns
	def getTableName(self):
		return "users_groups"
	def getKeys(self):
		return { 'user_id':'uuid', 'group_id':'uuid' }
	def getFK(self):
		return [ ForeignKey("user_id","users","id"), ForeignKey("group_id","groups","id") ]
	def getOrderBy(self):
		return [ "user_id", "group_id" ]
	def getDefaultEntries(self):
		return [\
			{'user_id':'-1','group_id':'-2',},\
			{'user_id':'-1','group_id':'-5',},\
			{'user_id':'-1','group_id':'-6',},\
		]
dbschema['users_groups'] = DBEUserGroup

class DBEGroup(DBEntity):
	__columns={}
	def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
		DBEntity.__init__(self, tablename, names, values, attrs, keys )
		if len(DBEGroup.__columns.keys())==0:
			parentcols = self.getColumns()
			for k in parentcols.keys():
				DBEGroup.__columns[k] = parentcols[k]
			DBEGroup.__columns['id'] = ["uuid","not null"]
			DBEGroup.__columns['name'] = ["varchar(255)","not null"]
			DBEGroup.__columns['description'] = ["text","default null"]
		self._columns = DBEGroup.__columns
	def getTableName(self):
		return "groups"
	def getKeys(self):
		return { 'id':'uuid' }
	def getFK(self):
		return [  ]
	def getOrderBy(self):
		return [ "name" ]
	def getDefaultEntries(self):
		return [\
			{'id':'-2','name':'Admin','description':'System admins'},\
			{'id':'-3','name':'Users','description':'System users'},\
			{'id':'-4','name':'Guests','description':'System guests (read only)'},\
			{'id':'-5','name':'Project','description':'R-Project user'},\
			{'id':'-6','name':'Webmaster','description':'Web content creators'},\
		]
	def _before_insert(self,dbmgr=None):
		if self.getValue('id') is None:
			myid = dbmgr.getNextUuid(self)
			self.setValue( 'id', myid )
	def _after_insert(self,dbmgr=None):
		if dbmgr.getDBEUser() is None:
			return
		dbe = DBEUserGroup()
		dbe.setValue('group_id',self.getValue('id'))
		dbe.setValue('user_id', dbmgr.getDBEUser().getValue('id'))
		dbmgr.insert(dbe)
		dbmgr.addGroup( self.getValue('id') )
	def _after_delete(self,dbmgr=None):
		cerca = DBEUserGroup()
		cerca.setValue('group_id', self.getValue('id'))
		lista = dbmgr.search(cerca, uselike=0)
		for ass in lista:
			dbmgr.delete(ass)
dbschema['groups'] = DBEGroup

class DBEUser(DBEntity):
	__columns={}
	def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
		DBEntity.__init__(self, tablename, names, values, attrs, keys )
		if len(DBEUser.__columns.keys())==0:
			parentcols = self.getColumns()
			for k in parentcols.keys():
				DBEUser.__columns[k] = parentcols[k]
			DBEUser.__columns['id'] = ["uuid","not null"]
			DBEUser.__columns['login'] = ["varchar(255)","not null"]
			DBEUser.__columns['pwd'] = ["varchar(255)","not null"]
			DBEUser.__columns['pwd_salt'] = ["varchar(4)","default ''"]
			DBEUser.__columns['fullname'] = ["text","default null"]
			DBEUser.__columns['group_id'] = ["uuid","not null"]
		self._columns=DBEUser.__columns
	def getValue( self, chiave ):
		if chiave=='name':
			return DBEntity.getValue( self, 'login' )
		return DBEntity.getValue( self, chiave )
	def setValue( self, chiave, valore ):
		if chiave=='name':
			return DBEntity.setValue( self, 'login', valore )
		return DBEntity.setValue( self, chiave, valore )
	def getTableName(self):
		return "users"
	def getKeys(self):
		return { 'id':'uuid' }
	def getFK(self):
		return [ ForeignKey("group_id","groups","id") ]
	def getOrderBy(self):
		return [ "fullname" ]
	def getDefaultEntries(self):
		return [\
			{'id':'-1', 'login':'adm','pwd':'adm','pwd_salt':'','fullname':'Administrator','group_id':'-2'},\
		]
	def checkNewPassword(self):
		ret = True
		return ret
	def _before_insert(self,dbmgr=None):
		if self.getValue('id') is None:
			myid = dbmgr.getNextUuid(self)
			self.setValue('id',myid)
		self.checkNewPassword()
		if self.checkNewPassword():
			self._createGroup(dbmgr)
	def _after_insert(self,dbmgr=None):
		self._checkGroupAssociation(dbmgr)
	def _after_update(self,dbmgr=None):
		self._checkGroupAssociation(dbmgr)
	def _after_delete(self,dbmgr=None):
		cerca = DBEUserGroup()
		cerca.setValue('user_id',self.getValue('id'))
		lista = dbmgr.search(cerca, uselike = False)
		for ass in lista:
			dbmgr.delete(ass)
		self._deleteGroup(dbmgr)
	def _createGroup(self,dbmgr):
		if not self.getValue('group_id') is None: return
		dbe = DBEGroup()
		dbe.setValue('name', self.getValue('login'))
		dbe.setValue('description', "Private group for %s-%s" % (self.getValue('id'),self.getValue('login')) )
		dbe = dbmgr.insert(dbe)
		self.setValue('group_id', dbe.getValue('id'))
	def _deleteGroup(self,dbmgr):
		dbe = DBEGroup()
		dbe.setValue('id', self.getValue('group_id'))
		dbe = dbmgr.delete(dbe)
	def _checkGroupAssociation(self,dbmgr):
		ug = DBEUserGroup()
		ug.setValue('user_id',self.getValue('id'))
		ug.setValue('group_id',self.getValue('group_id'))
		exists = dbmgr.exists(ug)
		if not exists:
			dbmgr.insert(ug)
	def isRoot(self):
		return self.getValue('id')==-1 or self.getValue('id')=='-1'
dbschema['users'] = DBEUser

class DBELog(DBEntity):
	__columns={}
	def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
		DBEntity.__init__(self, tablename, names, values, attrs, keys )
		if len(DBELog.__columns.keys())==0:
			parentcols = self.getColumns()
			for k in parentcols.keys():
				DBELog.__columns[k] = parentcols[k]
			DBELog.__columns['ip'] = ["varchar(16)","not null"]
			DBELog.__columns['data'] = ["date","not null default '0000-00-00'"]
			DBELog.__columns['ora'] = ["time","not null default '00:00:00'"]
			DBELog.__columns['count'] = ["int","not null default 0"]
			DBELog.__columns['url'] = ["varchar(255)","default null"]
			DBELog.__columns['note'] = ["varchar(255)","not null default ''"]
			DBELog.__columns['note2'] = ["text","not null"]
		self._columns=DBELog.__columns
	def getTableName(self):
		return "log"
	def getKeys(self):
		return { 'ip':'varchar(16)', 'data':'date' }
	def getFK(self):
		return [  ]
	def getOrderBy(self):
		return [ "data desc", "ora desc" ]
	def _before_insert(self,dbmgr=None):
		pass
dbschema['log'] = DBELog

class DBEObject(DBEntity):
	__columns={}
	def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
		DBEntity.__init__(self, tablename, names, values, attrs, keys )
		if len(DBEObject.__columns.keys())==0:
			parentcols=self.getColumns()
			for k in parentcols.keys():
				DBEObject.__columns[k]=parentcols[k]
			DBEObject.__columns['id']=["uuid","not null"]
			DBEObject.__columns['owner']=["uuid","not null"]
			DBEObject.__columns['group_id']=["uuid","not null"]
			DBEObject.__columns['permissions']=["char(9)","not null default 'rwx------'"]
			DBEObject.__columns['creator']=["uuid","not null"]
			DBEObject.__columns['creation_date']=["datetime","default null"]
			DBEObject.__columns['last_modify']=["uuid","not null"]
			DBEObject.__columns['last_modify_date']=["datetime","default null"]
			DBEObject.__columns['deleted_by']=["uuid","default null"]
			DBEObject.__columns['deleted_date']=["datetime","not null default '0000-00-00 00:00:00'"]
			DBEObject.__columns['father_id']=["uuid","default null"]
			DBEObject.__columns['name']=["varchar(255)","not null"]
			DBEObject.__columns['description']=["text","default null"]
		self._columns=DBEObject.__columns
	def getTableName(self):
		return "objects"
	def getKeys(self):
		return { 'id':'uuid' }
	def getFK(self):
		return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","objects","id") ]
	def getOrderBy(self):
		return [ "name" ]
	# ====== Object Methods ======
	def getOwnerId(self):
		return self.getValue('owner')
	def getGroupId(self):
		return self.getValue('group_id')
	def isDeleted(self):
		myvalue = self.getValue('deleted_date')
		if not myvalue is None and not type(myvalue)==str:
			myvalue = self._formatDateTime(myvalue)
		if myvalue>'0000-00-00 00:00:00':
			return True
		return False
	def canRead(self,kind=''):
		if not self.getValue('permissions')>'':
			return True
		elif kind=='U': # User
			tmp = self.getValue('permissions')[0+0:0+0+1]
			return tmp == 'r'
		elif kind=='G': # Group
			tmp = self.getValue('permissions')[0+3:0+3+1]
			return tmp == 'r'
		else: # All
			tmp = self.getValue('permissions')[0+6:0+6+1]
			return tmp == 'r'
	def canWrite(self,kind=''):
		if not self.getValue('permissions')>'':
			return True
		elif kind=='U': # User
			tmp = self.getValue('permissions')[1+0:1+0+1]
			return tmp == 'w'
		elif kind=='G': # Group
			tmp = self.getValue('permissions')[1+3:1+3+1]
			return tmp == 'w'
		else: # All
			tmp = self.getValue('permissions')[1+6:1+6+1]
			return tmp == 'w'
	def canExecute(self,kind=''):
		if not self.getValue('permissions')>'':
			return True
		elif kind=='U': # User
			tmp = self.getValue('permissions')[2+0:2+0+1]
			return tmp == 'x'
		elif kind=='G': # Group
			tmp = self.getValue('permissions')[2+3:2+3+1]
			return tmp == 'x'
		else: # All
			tmp = self.getValue('permissions')[2+6:2+6+1]
			return tmp == 'x'
	def setDefaultValues(self,dbmgr):
		myuser = dbmgr.getDBEUser()
		if not myuser is None:
			if not self.getValue('owner')>'':
				self.setValue('owner',myuser.getValue('id'))
			if not self.getValue('group_id')>'':
				self.setValue('group_id',myuser.getValue('group_id'))
			self.setValue('creator',myuser.getValue('id'))
			self.setValue('last_modify',myuser.getValue('id'))
		self.setValue('creation_date',self._getTodayString())
		self.setValue('last_modify_date',self._getTodayString())
		if dbmgr._connProvider.getDBType()=="POSTGRESQL":
			self.setValue('deleted_date',None)
		else:
			self.setValue('deleted_date','0000-00-00 00:00:00')
		
		if not self.getValue('permissions')>'':
			self.setValue('permissions','rwx------')
		
		father_id = self.getValue('father_id')
		if father_id is None or father_id==0 or len("%s"%father_id)==0:
			self.setValue('father_id','')
			
			if self.getValue('fk_obj_id')>'':
				fkobj = dbmgr.objectById( self.getValue('fk_obj_id') )
				self.setValue('group_id',fkobj.getValue('group_id'))
				self.setValue('permissions',fkobj.getValue('permissions'))
				self.setValue('father_id',self.getValue('fk_obj_id'))
		else:
			father = dbmgr.objectById( father_id )
			if not father is None:
				self.setValue('owner',father.getValue('owner'))
				self.setValue('group_id',father.getValue('group_id'))
				self.setValue('permissions',father.getValue('permissions'))
	def _before_insert(self, dbmgr):
		if self.getValue('id') is None:
			myid = dbmgr.getNextUuid(self)
			self.setValue( 'id', myid )
		self.setDefaultValues(dbmgr)
	def _before_update(self, dbmgr):
		myuser = dbmgr.getDBEUser()
		if not myuser is None:
			self.setValue('last_modify',myuser.getValue('id'))
		if dbmgr._connProvider.getDBType()=="POSTGRESQL" and self.getValue('deleted_date')=='0000-00-00 00:00:00':
			self.setValue('deleted_date',None)
		self.setValue('last_modify_date', self._getTodayString())
	def _before_delete(self, dbmgr):
		if self.isDeleted():
			return
		myuser = dbmgr.getDBEUser()
		if not myuser is None:
			self.setValue('deleted_by',myuser.getValue('id'))
		self.setValue('deleted_date', self._getTodayString())
dbschema['objects'] = DBEObject

class ObjectMgr(DBMgr):
    def __init__(self, connProvider, verbose=0, schema = ''):
        DBMgr.__init__(self, connProvider, verbose, schema)
    def canRead(self,obj):
        ret = False
        myuser = self.getDBEUser()
        if obj.canRead():
            ret = True
        elif obj.canRead('G') and self.hasGroup( obj.getGroupId() ):
            ret = True
        elif obj.canRead('U') and myuser is not None and myuser.getValue('id')==obj.getOwnerId():
            ret = True
        return ret
    def canWrite(self,obj):
        ret = False
        myuser = self.getDBEUser()
        if obj.canWrite():
            ret = True
        elif obj.canWrite('G') and self.hasGroup( obj.getGroupId() ):
            ret = True
        elif obj.canWrite('U') and myuser is not None and myuser.getValue('id')==obj.getOwnerId():
            ret = True
        return ret
    def canExecute(self,obj):
        ret = False
        myuser = self.getDBEUser()
        if obj.canExecute():
            ret = True
        elif obj.canExecute('G') and self.hasGroup( obj.getGroupId() ):
            ret = True
        elif obj.canExecute('U') and myuser is not None and myuser.getValue('id')==obj.getOwnerId():
            ret = True
        return ret
    def select(self,tablename,searchString):
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self.conn.select(tablename,searchString)
        if self._verbose: print "ObjectMgr.select: start."
        tmp = DBMgr.select(self,tablename,searchString)
        myuser = self.getDBEUser()
        if not myuser is None and myuser.isRoot():
            return tmp
        ret = []
        for obj in tmp:
            if isinstance(obj, DBEObject):
                if not myuser is None and obj.getValue('creator')==myuser.getValue('id'):
                    ret.append(obj)
                    continue
                if obj.isDeleted():
                    continue
                if self.canRead(obj):
                    ret.append(obj)
                continue
            elif isinstance(obj, DBEUser):
                if myuser is None or self.hasGroup(GROUP_ADMIN) or obj.getValue('id')==myuser.getValue('id'):
                    ret.append(obj)
                continue
            elif isinstance(obj, DBEGroup):
                if myuser is None or self.hasGroup(GROUP_ADMIN) or self.hasGroup( obj.getValue('id') ):
                    ret.append(obj)
                continue
            ret.append( obj )
        if self._verbose:
            print "ObjectMgr.select: ret=%s" % ( len( ret ) )
            print "ObjectMgr.select: end."
        return ret
    def insert(self, dbe, cleanKeysIfError=True):
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self.conn.insert(dbe)
        have_permission = True
        if isinstance(dbe, DBEObject):
            have_permission = False
            myuser = self.getDBEUser()
            if not myuser is None and dbe.getValue('creator')==myuser.getValue('id'):
                have_permission = True
            elif self.canWrite( dbe ):
                have_permission = True
        if have_permission:
            return DBMgr.insert(self, dbe, cleanKeysIfError)
        return None
    def update(self, dbe):
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self.conn.update(dbe)
        have_permission = True
        if isinstance(dbe, DBEObject):
            have_permission = False
            myuser = self.getDBEUser()
            if not myuser is None and dbe.getValue('creator')==myuser.getValue('id'):
                have_permission = True
            elif self.canWrite( dbe ):
                have_permission = True
        if have_permission:
            return DBMgr.update(self, dbe)
        return None
    def delete(self, dbe):
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self.conn.delete(dbe)
        have_permission = True
        if isinstance(dbe, DBEObject):
            have_permission = False
            myuser = self.getDBEUser()
            if not myuser is None and dbe.getValue('creator')==myuser.getValue('id'):
                have_permission = True
            elif self.canWrite( dbe ):
                have_permission = True
        if not have_permission:
            return dbe
        if not isinstance(dbe, DBEObject) or dbe.isDeleted():
            return DBMgr.delete(self, dbe)
        # Before Delete
        dbe._before_delete(self)
        dbe = self._before_delete(dbe)
        ## Delete
        query = self._buildUpdateString( dbe )
        if self._verbose: print "ObjectMgr.delete: query=%s" % query
        try:
            cursor = self.getConnection().cursor()
            cursor.execute( query )
            self.getConnection().commit()
            # After Delete
            dbe = self._after_delete(dbe)
            dbe._after_delete(self)
        except Exception, e:
            if self._verbose:
                print "ObjectMgr.delete: ECCEZIONE=%s (%s)" % (e,query)
                print "\n".join(traceback.format_tb(sys.exc_info()[2]))
            try:
                ris = self.getConnection().query( query )
                self.getConnection().commit()
                # After Delete
                dbe = self._after_delete(dbe)
                dbe._after_delete(self)
            except Exception, e:
                dbe = None
                raise DBLayerException("ObjectMgr.delete: unable to execute\n\t%s.Error cause: %s" % ( query, e ))
        return dbe
    def db_version(self):
        cerca = DBEDBVersion()
        ris = self.search(cerca,1)
        if len(ris)==1:
           return ris[0].version()
        else:
           return 0
    def _buildSelectString(self, dbe, uselike=1,case_sensitive=True):
        if not isinstance(dbe, DBEObject) or dbe.getTypeName()!="DBEObject":
            return DBMgr._buildSelectString(self, dbe, uselike, case_sensitive)
        tipi = self.getRegisteredTypes()
        q = []
        for tablename, clazz in tipi.items():
            mydbe = clazz()
            if not isinstance(mydbe, DBEObject) or isinstance(mydbe, DBAssociation)\
                    or mydbe.getTypeName()=="DBEObject":
                    #or isinstance(mydbe,DBETodoTipo) or mydbe.getTypeName()=="DBEObject":
                continue
            mydbe.setValuesDictionary(dbe.getValuesDictionary())
            q.append(
                DBMgr._buildSelectString(self, mydbe, uselike, case_sensitive).replace(
                    "select * ",
                    "select '%s' as classname,id,owner,group_id,permissions,creator,creation_date,last_modify,last_modify_date,father_id,name,description " % (mydbe.getTypeName())
                )
            )
        searchString = " union ".join( q )
        if self._verbose: print "ObjectMgr._buildSelectString: searchString=%s" % ( searchString )
        return searchString
    def search(self, dbe, uselike=1, orderby=None, ignore_deleted=True, full_object=True):
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self.conn.search(dbe, uselike, orderby, ignore_deleted, full_object)
        if self._verbose: print "ObjectMgr.search: start."
        if self._verbose: print "ObjectMgr.search: dbe.getTypeName()=%s." % (dbe.getTypeName())
        if dbe.getTypeName()!="DBEObject":
            if isinstance(dbe, DBEObject) and ignore_deleted:
                if self._connProvider.getDBType()=="POSTGRESQL":
                    dbe.setValue('deleted_date','null')
                else:
                    dbe.setValue('deleted_date','0000-00-00 00:00:00')
            ret = DBMgr.search(self, dbe, uselike, orderby)
            if self._verbose: print "ObjectMgr.search: quick end; ret = %s" % ( len(ret) )
            return ret
        if ignore_deleted:
            if self._connProvider.getDBType()=="POSTGRESQL":
                dbe.setValue('deleted_date','null')
            else:
                dbe.setValue('deleted_date','0000-00-00 00:00:00')
        ret = []
        tmp = DBMgr.search(self, dbe, uselike, orderby)
        if self._verbose:
            print "ObjectMgr.search: tmp=%s" % ( len(tmp) )
            for _t in tmp:
                print "ObjectMgr.search: -> %s" % ( _t )
        for _obj in tmp:
            cerca = self.getClazzByTypeName("%s"%_obj.getValue('classname'))()
            cerca.setValue('id',_obj.getValue('id'))
            lista = DBMgr.search(self,cerca,uselike=0)
            if self._verbose: print "ObjectMgr.search: lista=%s" % ( len(lista) )
            if len(lista)!=1:
                continue
            ret.append( lista[0] )
        if self._verbose: print "ObjectMgr.search: end."
        return ret
    def objectById(self,myid,ignore_deleted=True):
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self._connProvider.objectById(myid,ignore_deleted)
        tipi = self.getRegisteredTypes()
        q = []
        for tablename, clazz in tipi.items():
            mydbe = clazz()
            if not isinstance(mydbe, DBEObject) or isinstance(mydbe, DBAssociation): # or mydbe.getTypeName=="DBEObject":
                continue
            tmp_q = "select '%s' as classname,id,owner,group_id,permissions,creator,creation_date,last_modify,last_modify_date,father_id,name,description from %s where id='%s'" % (mydbe.getTypeName(), self.buildTableName(mydbe), DBEntity.hex2uuid(myid) )
            if ignore_deleted:
                if self._connProvider.getDBType()=="POSTGRESQL":
                    tmp_q += " and deleted_date is null"
                else:
                    tmp_q += " and deleted_date='0000-00-00 00:00:00'"
            q.append( tmp_q )
        searchString = " union ".join( q )
        if self._verbose: print "ObjectMgr.objectById: searchString=%s" % ( searchString )
        lista = self.select("objects", searchString)
        if len(lista)==1:
            return lista[0]
        if self._verbose: print "ObjectMgr.objectById: lista=%s" % ( len(lista) )
        return None
    def fullObjectById(self, myid,ignore_deleted=True):
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self._connProvider.fullObjectById(myid,ignore_deleted)
        myobj = self.objectById(myid,ignore_deleted)
        if myobj is None:
            return None
        cerca = self.getClazzByTypeName(myobj.getValue('classname'))()
        cerca.setValue('id',myobj.getValue('id'))
        lista = self.search(cerca,uselike=0,ignore_deleted=ignore_deleted)
        if self._verbose: print "ObjectMgr.fullObjectById: lista=%s" % ( len( lista ) )
        if len(lista)==1:
            return lista[0]
        return None
    def objectByName(self, name,ignore_deleted=True):
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self._connProvider.objectByName(name,ignore_deleted)
        tipi = self.getRegisteredTypes()
        q = []
        for tablename, clazz in tipi.items():
            mydbe = clazz()
            if not isinstance(mydbe, DBEObject) or isinstance(mydbe, DBAssociation): # or mydbe.getTypeName=="DBEObject":
                continue
            tmp_q = """
select '%s' as classname,id,owner,group_id,permissions,creator,
       creation_date,last_modify,last_modify_date,father_id,name,description
  from %s
 where name='%s'
""" % (mydbe.getTypeName(), self.buildTableName(mydbe), name )
            if ignore_deleted:
                if self._connProvider.getDBType()=="POSTGRESQL":
                    tmp_q += " and deleted_date is null"
                else:
                    tmp_q += " and deleted_date='0000-00-00 00:00:00'"
            q.append( tmp_q )
        searchString = " union ".join( q )
        if self._verbose: print "ObjectMgr.objectById: searchString=%s" % ( searchString )
        lista = self.select("objects", searchString)
        if len(lista)==1:
            return lista[0]
        return None
    def fullObjectByName(self, name, ignore_deleted = True):
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self._connProvider.fullObjectByName(name,ignore_deleted)
        myobj = self.objectByName(name,ignore_deleted)
        if myobj is None:
            return None
        cerca = self.getClazzByTypeName(myobj.getValue('classname'))()
        cerca.setValue('id',myobj.getValue('id'))
        lista = self.search(cerca,uselike=0,ignore_deleted=ignore_deleted)
        if self._verbose: print "ObjectMgr.fullObjectById: lista=%s" % ( len( lista ) )
        if len(lista)==1:
            return lista[0]
        return None
    def downloadFile(self,remote_filename,local_path,view_thumbnail=False):
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self._connProvider.downloadFile(remote_filename,local_path,view_thumbnail)
        myobj = self.fullObjectById(remote_filename)
        myobj._dest_directory = os.path.realpath(self.getConnectionProvider().getLocalFilePath())
        fullpath = myobj.getFullpath()
        local_filename = "%s%s%s" % (local_path,os.path.sep,myobj.getValue('name'))
        return self._connProvider.downloadFile(fullpath,local_filename,view_thumbnail)


class DBECountry(DBEntity):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEntity.__init__(self, tablename, names, values, attrs, keys )
        if len(DBECountry.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBECountry.__columns[k]=parentcols[k]
            DBECountry.__columns['id']=["uuid","not null"]
            DBECountry.__columns['Common_Name']=["varchar(255)","default null"]
            DBECountry.__columns['Formal_Name']=["varchar(255)","default null"]
            DBECountry.__columns['Type']=["varchar(255)","default null"]
            DBECountry.__columns['Sub_Type']=["varchar(255)","default null"]
            DBECountry.__columns['Sovereignty']=["varchar(255)","default null"]
            DBECountry.__columns['Capital']=["varchar(255)","default null"]
            DBECountry.__columns['ISO_4217_Currency_Code']=["varchar(255)","default null"]
            DBECountry.__columns['ISO_4217_Currency_Name']=["varchar(255)","default null"]
            DBECountry.__columns['ITU_T_Telephone_Code']=["varchar(255)","default null"]
            DBECountry.__columns['ISO_3166_1_2_Letter_Code']=["varchar(255)","default null"]
            DBECountry.__columns['ISO_3166_1_3_Letter_Code']=["varchar(255)","default null"]
            DBECountry.__columns['ISO_3166_1_Number']=["varchar(255)","default null"]
            DBECountry.__columns['IANA_Country_Code_TLD']=["varchar(255)","default null"]
        self._columns=DBECountry.__columns
    def getTableName(self):
        return "countrylist"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [  ]
    def getOrderBy(self):
        return [ "id" ]
dbschema['countrylist'] = DBECountry

class DBECompany(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBECompany.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBECompany.__columns[k]=parentcols[k]
            DBECompany.__columns['street']=["varchar(255)","default null"]
            DBECompany.__columns['zip']=["varchar(255)","default null"]
            DBECompany.__columns['city']=["varchar(255)","default null"]
            DBECompany.__columns['state']=["varchar(255)","default null"]
            DBECompany.__columns['fk_countrylist_id']=["uuid","default null"]
            DBECompany.__columns['phone']=["varchar(255)","default null"]
            DBECompany.__columns['fax']=["varchar(255)","default null"]
            DBECompany.__columns['email']=["varchar(255)","default null"]
            DBECompany.__columns['url']=["varchar(255)","default null"]
            DBECompany.__columns['p_iva']=["varchar(16)","default null"]
        self._columns=DBECompany.__columns
    def getTableName(self):
        return "companies"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","companies","id"), ForeignKey("fk_countrylist_id","countrylist","id") ]
    def getOrderBy(self):
        return [ "name" ]
dbschema['companies'] = DBECompany

class DBEPeople(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEPeople.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEPeople.__columns[k]=parentcols[k]
            DBEPeople.__columns['street']=["varchar(255)","default null"]
            DBEPeople.__columns['zip']=["varchar(255)","default null"]
            DBEPeople.__columns['city']=["varchar(255)","default null"]
            DBEPeople.__columns['state']=["varchar(255)","default null"]
            DBEPeople.__columns['fk_countrylist_id']=["uuid","default null"]
            DBEPeople.__columns['fk_companies_id']=["uuid","default null"]
            DBEPeople.__columns['fk_users_id']=["uuid","default null"]
            DBEPeople.__columns['phone']=["varchar(255)","default null"]
            DBEPeople.__columns['office_phone']=["varchar(255)","default null"]
            DBEPeople.__columns['mobile']=["varchar(255)","default null"]
            DBEPeople.__columns['fax']=["varchar(255)","default null"]
            DBEPeople.__columns['email']=["varchar(255)","default null"]
            DBEPeople.__columns['url']=["varchar(255)","default null"]
            DBEPeople.__columns['codice_fiscale']=["varchar(20)","default null"]
            DBEPeople.__columns['p_iva']=["varchar(16)","default null"]
        self._columns=DBEPeople.__columns
    def getTableName(self):
        return "people"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","people","id"), ForeignKey("fk_countrylist_id","countrylist","id"), ForeignKey("fk_companies_id","companies","id"), ForeignKey("fk_users_id","users","id") ]
    def getOrderBy(self):
        return [ "name" ]
dbschema['people'] = DBEPeople

class DBEEvent(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEEvent.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEEvent.__columns[k]=parentcols[k]
            DBEEvent.__columns['fk_obj_id']=["uuid","default null"]
            DBEEvent.__columns['start_date']=["datetime","not null default '0000-00-00 00:00:00'"]
            DBEEvent.__columns['end_date']=["datetime","not null default '0000-00-00 00:00:00'"]
            DBEEvent.__columns['all_day']=["char(1)","not null default '1'"]
            DBEEvent.__columns['url']=["varchar(255)","default null"]
            DBEEvent.__columns['alarm']=["char(1)","default '0'"]
            DBEEvent.__columns['alarm_minute']=["int","default 0"]
            DBEEvent.__columns['alarm_unit']=["char(1)","default '0'"]
            DBEEvent.__columns['before_event']=["char(1)","default '0'"]
            DBEEvent.__columns['category']=["varchar(255)","default ''"]
            DBEEvent.__columns['recurrence']=["char(1)","default '0'"]
            DBEEvent.__columns['recurrence_type']=["char(1)","default '0'"]
            DBEEvent.__columns['daily_every_x']=["int","default 0"]
            DBEEvent.__columns['weekly_every_x']=["int","default 0"]
            DBEEvent.__columns['weekly_day_of_the_week']=["char(1)","default '0'"]
            DBEEvent.__columns['monthly_every_x']=["int","default 0"]
            DBEEvent.__columns['monthly_day_of_the_month']=["int","default 0"]
            DBEEvent.__columns['monthly_week_number']=["int","default 0"]
            DBEEvent.__columns['monthly_week_day']=["char(1)","default '0'"]
            DBEEvent.__columns['yearly_month_number']=["int","default 0"]
            DBEEvent.__columns['yearly_month_day']=["int","default 0"]
            DBEEvent.__columns['yearly_week_number']=["int","default 0"]
            DBEEvent.__columns['yearly_week_day']=["char(1)","default '0'"]
            DBEEvent.__columns['yearly_day_of_the_year']=["int","default 0"]
            DBEEvent.__columns['recurrence_times']=["int","default 0"]
            DBEEvent.__columns['recurrence_end_date']=["datetime","not null default '0000-00-00 00:00:00'"]
        self._columns=DBEEvent.__columns
    def getTableName(self):
        return "events"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","events","id"), ForeignKey("fk_obj_id","companies","id"), ForeignKey("fk_obj_id","folders","id"), ForeignKey("fk_obj_id","people","id"), ForeignKey("fk_obj_id","projects","id") ]
    def getOrderBy(self):
        return [ "name" ]
dbschema['events'] = DBEEvent

class DBEFile(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEFile.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEFile.__columns[k]=parentcols[k]
            DBEFile.__columns['fk_obj_id']=["uuid","default null"]
            DBEFile.__columns['path']=["text","default null"]
            DBEFile.__columns['filename']=["text","not null"]
            DBEFile.__columns['checksum']=["char(40)","default null"]
            DBEFile.__columns['mime']=["varchar(255)","default null"]
            DBEFile.__columns['alt_link']=["varchar(255)","not null default ''"]
        self._columns=DBEFile.__columns
        
        self._dest_directory = ''
    def getTableName(self):
        return "files"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","files","id"), ForeignKey("father_id","folders","id"), ForeignKey("fk_obj_id","pages","id"), ForeignKey("father_id","pages","id"), ForeignKey("fk_obj_id","news","id"), ForeignKey("father_id","news","id") ]
    def getOrderBy(self):
        return [ "path", "filename" ]
    def generaObjectPath(self,a_dbe=None):
        _dbe = self
        if not a_dbe is None:
            _dbe = a_dbe
        dest_path = ''
        if _dbe.getValue('path')>'':
            dest_path = _dbe.getValue('path')
        father_id = _dbe.getValue('father_id')
        if father_id>'':
            if dest_path>'':
                dest_path = "%s%s%s" % (father_id,os.path.sep,dest_path)
            else:
                dest_path = "%s" % (father_id)
        return dest_path
    def getFullpath(self,a_dest_dir="",a_dbe=None):
        _dbe = self
        if not a_dbe is None:
            _dbe = a_dbe
        dest_path = _dbe.generaObjectPath()
        dest_dir = a_dest_dir
        if a_dest_dir=='':
            dest_dir = self._dest_directory
        if dest_path>'': dest_dir="%s%s%s" % (dest_dir,os.path.sep,dest_path)
        ret = dest_dir="%s%s%s" % (dest_dir,os.path.sep,_dbe.getValue('filename'))
        return ret
    def generaFilename(self, aId=None, aFilename=None):
        nomefile = self.getValue('filename')
        if not aFilename is None:
            nomefile = aFilename
        _id=self.getValue('id')
        if not aId is None:
            _id = aId
        prefisso = "r_%s_" % _id
        if nomefile.find(prefisso)>=0:
            nomefile = nomefile.replace(prefisso,"")
        return "%s%s" % (prefisso,nomefile)
    # Image management: start.
    def getThumbnailFilename(self):
        return "%s_thumb.jpg" % (self.getValue('filename'))
    def isImage(self):
        _mime = self.getValue('mime')
        return _mime>'' and _mime[0:5]=='image'
    # TODO
    def createThumbnail(self,fullpath,pix_width=100,pix_height=100):
        try:
            # Please install Python Imaging :-)
            from PIL import Image
            image = Image.open(fullpath)
            image = image.resize((pix_width, pix_height), Image.ANTIALIAS)
            image.save("%s_thumb.jpg" % fullpath)
            return "%s_thumb.jpg" % fullpath
        except Exception,e:
            return ""
    def deleteThumbnail(self,fullpath):
        os.remove("%s_thumb.jpg" % fullpath)
    # Image management: end.
    
    def _before_insert(self,dbmgr=None):
        DBEObject._before_insert(self,dbmgr)
        self._dest_directory = os.path.realpath(dbmgr.getConnectionProvider().getLocalFilePath())
        
        if not self.getValue('filename')>'' or not os.path.exists(self.getValue('filename')):
            return
        if self.getValue('name')=='New DBEFile':
            self.setValue('name','')
        
        if self.getValue('filename')>'':
            # Uploading
            local_filename = self.getValue('filename')
            self.setValue('filename', dbmgr.uploadFile(self.getValue('filename')) )
            
            dest_path = self.generaObjectPath()
            from_dir = os.path.realpath(dbmgr.getConnectionProvider().getLocalFilePath())
            dest_dir = os.path.realpath(dbmgr.getConnectionProvider().getLocalFilePath())
            if dest_path>'':
                dest_dir = "%s%s%s" % (dest_dir,os.path.sep,dest_path)
            if not os.path.exists(dest_dir): os.mkdir(dest_dir)
            nuovo_filename = self.generaFilename(\
                    self.getValue('id'),\
                    self.getValue('filename').split(os.path.sep)[-1]\
                    )
            filefrom=self.getValue('filename')
            fileto="%s%s%s" % (dest_dir,os.path.sep,nuovo_filename )
            os.rename( filefrom, fileto )
            if not self.getValue('name')>'':
                self.setValue('name',self.getValue('filename').split(os.path.sep)[-1])
            self.setValue('filename', nuovo_filename)
        # Checksum
        _fullpath = self.getFullpath()
        if os.path.exists(_fullpath):
            f = open(_fullpath,'rb')
            import hashlib
            h = hashlib.sha1()
            h.update(f.read())
            newchecksum = h.hexdigest()
            f.close()
            self.setValue('checksum',newchecksum)
        else:
            self.setValue('checksum',"File '%s' not found!"%self.getValue('filename'))
        # Mime type
        mime = 'text/plain'
        if os.path.exists(_fullpath):
            import mimetypes
            mime,altro = mimetypes.guess_type("file://%s"%_fullpath)
        self.setValue('mime',mime)
        # Image
        if self.isImage():
            thumbpath = self.createThumbnail(_fullpath)
    def _before_update(self,dbmgr=None):
        DBEObject._before_update(self,dbmgr)
        self._dest_directory = os.path.realpath(dbmgr.getConnectionProvider().getLocalFilePath())
        # Inherits father' root
        father_id = self.getValue('father_id')
        if father_id>'':
            query="select fk_obj_id from %s  where id='%s'" % (dbmgr.buildTableName(self),self.getValue('father_id'))
            tmp = dbmgr.select(self.getTableName(),query)
            if len(tmp)==1:
                self.setValue('fk_obj_id', tmp[0].getValue('fk_obj_id') )
        # An old file exists?
        cerca = dbmgr.getClazzByTypeName(self.getTypeName())()
        cerca.setValue('id',self.getValue('id'))
        lista = dbmgr.search(cerca,uselike=0)
        myself = lista[0]
        if self.getValue('filename')>'' and myself.getValue('filename')!=self.getValue('filename')\
                and os.path.exists(self.getValue('filename')):
            # Different filenames ==> remove the old file
            dest_path = myself.generaObjectPath()
            dest_dir = os.path.realpath(dbmgr.getConnectionProvider().getLocalFilePath())
            if dest_path>'':
                dest_dir = "%s%s%s" % (dest_dir,os.path.sep,dest_path)
            dest_file = "%s%s%s" % (dest_dir,os.path.sep,myself.generaFilename())
            if os.path.exists(dest_file):
                os.remove(dest_file)
                if self.isImage():
                    self.deleteThumbnail(dest_file)
        
        if self.getValue('filename')>'':
            # Computing from_*
            from_dir = os.path.realpath(dbmgr.getConnectionProvider().getLocalFilePath())
            from_path = self.generaObjectPath(myself)
            from_filename = self.generaFilename(\
                    self.getValue('id'),\
                    self.getValue('filename').split(os.path.sep)[-1]\
                    )
            if from_path>'':
                # FIXME why this works?
                from_dir = "%s%s%s" % (from_dir,os.path.sep,from_path)
                #from_dir = os.path.realpath(dbmgr.getConnectionProvider().getLocalFilePath())
                #filefrom=self.getValue('filename')
            # Uploading
            if os.path.exists(self.getValue('filename')):
                local_filename = self.getValue('filename')
                self.setValue('filename', dbmgr.uploadFile(self.getValue('filename')) )
                from_filename = self.getValue('filename')
                filefrom = from_filename
            else:
                filefrom="%s%s%s" % (from_dir,os.path.sep,from_filename)
            dest_path = self.generaObjectPath()
            dest_dir = os.path.realpath(dbmgr.getConnectionProvider().getLocalFilePath())
            if dest_path>'':
                dest_dir = "%s%s%s" % (dest_dir,os.path.sep,dest_path)
            if not os.path.exists(dest_dir): os.mkdir(dest_dir)
            nuovo_filename = self.generaFilename(\
                    self.getValue('id'),\
                    self.getValue('filename').split(os.path.sep)[-1]\
                    )
            fileto="%s%s%s" % (dest_dir,os.path.sep,nuovo_filename)
            os.rename( filefrom, fileto )
            #raise DBLayerException("File '%s' does not exists"%filefrom)
            if not os.path.exists(fileto):
                raise DBLayerException("File '%s' does not exists"%fileto)
            if not self.getValue('name')>'':
                self.setValue('name',self.getValue('filename').split(os.path.sep)[-1])
            self.setValue('filename', nuovo_filename)
        elif myself.getValue('path')!=self.getValue('path'):
            from_path = self.generaObjectPath()
            from_dir = os.path.realpath(dbmgr.getConnectionProvider().getLocalFilePath())
            if from_path>'':
                from_dir = "%s%s%s" % (from_dir,os.path.sep,from_path)
            dest_path = self.generaObjectPath()
            dest_dir = os.path.realpath(dbmgr.getConnectionProvider().getLocalFilePath())
            if dest_path>'':
                dest_dir = "%s%s%s" % (dest_dir,os.path.sep,dest_path)
            if not os.path.exists(dest_dir): os.mkdir(dest_dir)
            os.rename( "%s%s%s" % (from_dir,os.path.sep,myself.getValue('filename')),\
                        "%s%s%s" % (dest_dir,os.path.sep,myself.getValue('filename') ) )
            self.setValue('filename', myself.getValue('filename'))
        else:
            self.setValue('filename', myself.getValue('filename'))
        # Checksum
        _fullpath = self.getFullpath()
        if os.path.exists(_fullpath):
            f = open(_fullpath,'rb')
            import hashlib
            h = hashlib.sha1() 
            h.update(f.read()) 
            newchecksum = h.hexdigest() 
            f.close()
            self.setValue('checksum',newchecksum)
        else:
            self.setValue('checksum',"File '%s' not found!"%self.getValue('filename'))
        # Mime type
        mime = 'text/plain'
        if os.path.exists(_fullpath):
            import mimetypes
            mime,altro = mimetypes.guess_type("file://%s"%_fullpath)
        self.setValue('mime',mime)
        # Image
        if self.isImage():
            thumbpath = self.createThumbnail(_fullpath)
    def _before_delete(self,dbmgr=None):
        is_deleted = self.isDeleted()
        DBEObject._before_delete(self,dbmgr)
        self._dest_directory = os.path.realpath(dbmgr.getConnectionProvider().getLocalFilePath())
        if is_deleted:
            # An old file exists?
            cerca = dbmgr.getClazzByTypeName(self.getTypeName())()
            cerca.setValue('id',self.getValue('id'))
            tmp = dbmgr.search(cerca,0,None,False)#,False,None,False)
            if len(tmp)>0:
                myself=tmp[0]
                if myself.getValue('filename')>'':
                    # ==> remove file
                    dest_path = self.generaObjectPath()
                    dest_dir = os.path.realpath(dbmgr.getConnectionProvider().getLocalFilePath())
                    if dest_path>'':
                        dest_dir = "%s%s%s" % (dest_dir,os.path.sep,dest_path)
                    os.remove( "%s%s%s" % (dest_dir,os.path.sep,myself.generaFilename() ) )
                    # Image
                    if self.isImage():
                        self.deleteThumbnail("%s%s%s" % (dest_dir,os.path.sep,myself.generaFilename() ))
dbschema['files'] = DBEFile

class DBEFolder(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEFolder.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEFolder.__columns[k]=parentcols[k]
            DBEFolder.__columns['fk_obj_id']=["uuid","default null"]
            DBEFolder.__columns['childs_sort_order']=["text","default null"]
        self._columns=DBEFolder.__columns
    def getTableName(self):
        return "folders"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","folders","id"), ForeignKey("fk_obj_id","companies","id"), ForeignKey("fk_obj_id","people","id"), ForeignKey("fk_obj_id","projects","id") ]
    def getOrderBy(self):
        return [ "name" ]
    def setDefaultValues(self,dbmgr):
        DBEObject.setDefaultValues(self,dbmgr)
        
        if self.getValue('father_id') is None:
            pass
        else:
            cerca = DBEFolder()
            cerca.setValue('id',self.getValue('father_id'))
            lista = dbmgr.search(cerca,uselike=0)
            if len(lista)==1:
                father = lista[0]
                self.setValue('fk_obj_id',father.getValue('fk_obj_id'))
    def _before_insert(self,dbmgr=None):
        DBEObject._before_insert(self,dbmgr)
        father_id = self.getValue('father_id')
        if father_id>'':
            query = "select fk_obj_id from %s where id='%s'"%(\
                dbmgr.buildTableName(self),\
                ("%-16s"%self.getValue('father_id')).replace(' ','\0')\
                )
            tmp = dbmgr.select(self.getTableName(),query)
            if len(tmp)==1:
                self.setValue('fk_obj_id', tmp[0].getValue('fk_obj_id') )
    def _before_update(self,dbmgr=None):
        DBEObject._before_update(self,dbmgr)
        father_id = self.getValue('father_id')
        if father_id>'':
            query = "select fk_obj_id from %s where id='%s'"%(\
                dbmgr.buildTableName(self),\
                ("%-16s"%self.getValue('father_id')).replace(' ','\0')\
                )
            tmp = dbmgr.select(self.getTableName(),query)
            if len(tmp)==1:
                self.setValue('fk_obj_id', tmp[0].getValue('fk_obj_id') )
    def _before_delete(self,dbmgr=None):
        DBEObject._before_delete(self,dbmgr)
        cerca = DBEObject()
        cerca.setValue('father_id',self.getValue('id'))
        lista = dbmgr.search(cerca,uselike=0)
        for l in lista:
            dbmgr.delete(l)
dbschema['folders'] = DBEFolder

class DBELink(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBELink.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBELink.__columns[k]=parentcols[k]
            DBELink.__columns['href']=["varchar(255)","not null"]
            DBELink.__columns['target']=["varchar(255)","default '_blank'"]
            DBELink.__columns['fk_obj_id']=["uuid","default null"]
        self._columns=DBELink.__columns
    def getTableName(self):
        return "links"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","links","id"), ForeignKey("fk_obj_id","companies","id"), ForeignKey("fk_obj_id","folders","id"), ForeignKey("fk_obj_id","people","id"), ForeignKey("fk_obj_id","projects","id"), ForeignKey("fk_obj_id","pages","id"), ForeignKey("father_id","pages","id"), ForeignKey("fk_obj_id","news","id"), ForeignKey("father_id","news","id") ]
    def getOrderBy(self):
        return [ "name" ]
dbschema['links'] = DBELink

class DBENote(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBENote.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBENote.__columns[k]=parentcols[k]
            DBENote.__columns['fk_obj_id']=["uuid","default null"]
        self._columns=DBENote.__columns
    def getTableName(self):
        return "notes"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","notes","id"), ForeignKey("fk_obj_id","companies","id"), ForeignKey("fk_obj_id","folders","id"), ForeignKey("fk_obj_id","people","id"), ForeignKey("fk_obj_id","projects","id") ]
    def getOrderBy(self):
        return [ "name" ]
dbschema['notes'] = DBENote

class DBEPage(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEPage.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEPage.__columns[k]=parentcols[k]
            DBEPage.__columns['html']=["text","default null"]
            DBEPage.__columns['fk_obj_id']=["uuid","default null"]
            DBEPage.__columns['language']=["varchar(5)","default 'en_us'"]
        self._columns=DBEPage.__columns
    def getTableName(self):
        return "pages"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","pages","id"), ForeignKey("fk_obj_id","companies","id"), ForeignKey("fk_obj_id","folders","id"), ForeignKey("fk_obj_id","people","id"), ForeignKey("fk_obj_id","projects","id") ]
    def getOrderBy(self):
        return [ "name" ]
dbschema['pages'] = DBEPage

class DBENews(DBEPage):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEPage.__init__(self, tablename, names, values, attrs, keys )
        if len(DBENews.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBENews.__columns[k]=parentcols[k]
        self._columns=DBENews.__columns
    def getTableName(self):
        return "news"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","news","id"), ForeignKey("fk_obj_id","companies","id"), ForeignKey("fk_obj_id","folders","id"), ForeignKey("fk_obj_id","people","id"), ForeignKey("fk_obj_id","projects","id") ]
    def getOrderBy(self):
        return [ "name" ]
dbschema['news'] = DBENews

class DBEBanned(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEBanned.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEBanned.__columns[k]=parentcols[k]
            DBEBanned.__columns['ban_ip']=["varchar(40)","default null"]
            DBEBanned.__columns['redirected_to']=["varchar(40)","default 'http://adf.ly/XdZw'"]
            DBEBanned.__columns['give_reason']=["varchar(255)","default 'Your IP has been blocked.<br/>See http://adf.ly/XdZw for more details.'"]
        self._columns=DBEBanned.__columns
    def getTableName(self):
        return "banned"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","banned","id") ]
    def getOrderBy(self):
        return [ "creation_date desc, ban_ip" ]
    def _before_insert(self,dbmgr=None):
        DBEObject._before_insert(self,dbmgr)
        self.setValue( 'permissions', 'rwxr--r--' )
dbschema['banned'] = DBEBanned

class DBEMail(DBEFile):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEFile.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEMail.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEMail.__columns[k]=parentcols[k]
            DBEMail.__columns['msgid']=["varchar(255)","not null default ''"]
            DBEMail.__columns['subject']=["varchar(255)","default null"]
            DBEMail.__columns['msgfrom']=["varchar(255)","default null"]
            DBEMail.__columns['msgto']=["varchar(255)","default null"]
            DBEMail.__columns['msgcc']=["varchar(255)","default null"]
            DBEMail.__columns['msgdate']=["datetime","default '1000-01-01 00:00:00'"]
            DBEMail.__columns['msgbody']=["text","default ''"]
        self._columns=DBEMail.__columns
    def getTableName(self):
        return "mail"
    def getKeys(self):
        return { 'id':'uuid',  'msgid':'string' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","mail","id"), ForeignKey("father_id","folders","id"), ForeignKey("fk_obj_id","pages","id"), ForeignKey("father_id","pages","id"), ForeignKey("fk_obj_id","news","id"), ForeignKey("father_id","news","id") ]
    def getOrderBy(self):
        return [ "msgdate desc" ]
    def _before_insert(self,dbmgr=None):
        DBEFile._before_insert(self,dbmgr)
        self.setValue( 'permissions', 'rwx------' )
        _fullpath = self.getFullpath()
        self.parseEmail(_fullpath)
        self.checkUniqueMessageId( self.getValue('msgid'), dbmgr)
    def checkUniqueMessageId( self,  msgid, dbmgr ):
        if msgid is None or len(msgid)==0:
            return
        cerca = DBEMail()
        cerca.setValue('msgid', msgid)
        lista = dbmgr.search(cerca, uselike=0)
        if len(lista)>0:
            raise DBLayerException("More than one email with the same Message-ID: '%s'"%msgid)
    def _before_update(self,dbmgr=None):
        DBEFile._before_update(self,dbmgr)
        _fullpath = self.getFullpath()
        self.parseEmail(_fullpath)
    def parseEmail(self,filename):
        f = file(filename,'rb').read()
        import email
        msg = email.message_from_string(f)
        if msg.has_key('Message-ID'):
            self.setValue('msgid', self._filterString(msg['Message-ID']) )
        if msg.has_key('Subject'):
            self.setValue('subject', self._filterString(msg['Subject']) )
            self.setValue('name', self._filterString(msg['Subject']) )
        if msg.has_key('From'):
            self.setValue('msgfrom', self._filterString(msg['From']) )
        if msg.has_key('To'):
            self.setValue('msgto', self._filterString(msg['To']) )
        if msg.has_key('Cc'):
            self.setValue('msgcc', self._filterString(msg['Cc']) )
        if msg.has_key('Date'):
            try:
                mytuple = email.utils.parsedate("%s"%msg['Date'])
                self.setValue('msgdate',datetime.datetime( mytuple[0],mytuple[1],mytuple[2],mytuple[3],mytuple[4],mytuple[5],mytuple[6] ))
            except Exception,e:
                pass
        msgbody = ''
        for part in msg.walk():
            if part.get_content_type()=='text/plain':
                msgbody = part.get_payload()
                if len(msgbody.strip())>0:
                    break
        self.setValue('msgbody', self._filterString(msgbody) )
    def _filterString(self, s):
        """See http://docs.python.org/howto/unicode.html"""
        ret = unicode(s, errors='replace') # replace with \uXXXX 
        return ret
dbschema['mail'] = DBEMail


# #### Projects: start.

class DBEProject(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEProject.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEProject.__columns[k]=parentcols[k]
        self._columns=DBEProject.__columns
    def getTableName(self):
        return "projects"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","projects","id") ]
    def getOrderBy(self):
        return [ "name" ]
    def _before_delete(self,dbmgr=None):
        DBEObject._before_delete(self,dbmgr)
dbschema['projects'] = DBEProject

class DBEProjectCompanyRole(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEProjectCompanyRole.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEProjectCompanyRole.__columns[k]=parentcols[k]
            DBEProjectCompanyRole.__columns['order_position']=["int","default 0"]
        self._columns=DBEProjectCompanyRole.__columns
    def getTableName(self):
        return "projects_companies_roles"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","projects_companies_roles","id") ]
    def getOrderBy(self):
        return [ "order_position", "id" ]
    def _before_insert(self,dbmgr=None):
        DBEObject._before_insert(self,dbmgr)
        query="select max(order_position) as order_position from %s" % dbmgr.buildTableName(self)
        tmp = dbmgr.select(self.getTableName(),query)
        if len(tmp)==1:
            tmp_position = tmp[0].getValue('order_position')
            if tmp_position is None:
                tmp_position = 0
            else:
                tmp_position = int(tmp_position)
            self.setValue('order_position', tmp_position + 1 )
        else:
            self.setValue('order_position', 1 )
dbschema['projects_companies_roles'] = DBEProjectCompanyRole

class DBEProjectCompany(DBEntity):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEntity.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEProjectCompany.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEProjectCompany.__columns[k]=parentcols[k]
            DBEProjectCompany.__columns['project_id']=["uuid","not null default ''"]
            DBEProjectCompany.__columns['company_id']=["uuid","not null default ''"]
            DBEProjectCompany.__columns['projects_companies_role_id']=["uuid","not null default ''"]
        self._columns=DBEProjectCompany.__columns
    def getTableName(self):
        return "projects_companies"
    def getKeys(self):
        return { 'project_id':'uuid', 'company_id':'uuid', 'projects_companies_role_id':'uuid' }
    def getFK(self):
        return [ ForeignKey("project_id","projects","id"), ForeignKey("company_id","companies","id"), ForeignKey("projects_companies_role_id","projects_companies_roles","id") ]
    def getOrderBy(self):
        return [ "project_id", "company_id", "projects_companies_role_id" ]
dbschema['projects_companies'] = DBEProjectCompany

class DBEProjectPeopleRole(DBEObject,DBAssociation): #DBEntity):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEProjectPeopleRole.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEProjectPeopleRole.__columns[k]=parentcols[k]
            DBEProjectPeopleRole.__columns['order_position']=["int","default 0"]
        self._columns=DBEProjectPeopleRole.__columns
    def getTableName(self):
        return "projects_people_roles"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [  ]
    def getOrderBy(self):
        return [ "order_position", "id" ]
    def getDefaultEntries(self):
        return [\
            {'order_position':'1','name':'Project Manager','owner':'-1','group_id':'-2','creator':'-1','last_modify':'-1',},\
            {'order_position':'2','name':'Analyst','owner':'-1','group_id':'-2','creator':'-1','last_modify':'-1',},\
            {'order_position':'3','name':'Designer','owner':'-1','group_id':'-2','creator':'-1','last_modify':'-1',},\
            {'order_position':'4','name':'Developer','owner':'-1','group_id':'-2','creator':'-1','last_modify':'-1',},\
            {'order_position':'5','name':'Graphical Designer','owner':'-1','group_id':'-2','creator':'-1','last_modify':'-1',},\
            ]
    def _before_insert(self,dbmgr=None):
        DBEObject._before_insert(self,dbmgr)
        tmp = []
        query="select max(order_position) as order_position from %s" % dbmgr.buildTableName(self)
        try:
            tmp = dbmgr.select(self.getTableName(),query)
        except Exception,  e:
            # TODO Likely we are on MongoDB
            pass
        if len(tmp)==1:
            tmp_position = tmp[0].getValue('order_position')
            if tmp_position is None:
                tmp_position = 0
            else:
                tmp_position = int(tmp_position)
            self.setValue('order_position', tmp_position + 1 )
        else:
            self.setValue('order_position', 1 )
dbschema['projects_people_roles'] = DBEProjectPeopleRole

class DBEProjectPeople(DBEntity):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEntity.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEProjectPeople.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEProjectPeople.__columns[k]=parentcols[k]
            DBEProjectPeople.__columns['project_id']=["uuid","not null default ''"]
            DBEProjectPeople.__columns['people_id']=["uuid","not null default ''"]
            DBEProjectPeople.__columns['projects_people_role_id']=["uuid","not null default ''"]
        self._columns=DBEProjectPeople.__columns
    def getTableName(self):
        return "projects_people"
    def getKeys(self):
        return { 'project_id':'uuid', 'people_id':'uuid', 'projects_people_role_id':'uuid' }
    def getFK(self):
        return [ ForeignKey("project_id","projects","id"), ForeignKey("people_id","people","id"), ForeignKey("projects_people_role_id","projects_people_roles","id") ]
    def getOrderBy(self):
        return [ "project_id", "people_id", "projects_people_role_id" ]
dbschema['projects_people'] = DBEProjectPeople

class DBEProjectProjectRole(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEProjectProjectRole.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEProjectProjectRole.__columns[k]=parentcols[k]
            DBEProjectProjectRole.__columns['order_position']=["int","default 0"]
        self._columns=DBEProjectProjectRole.__columns
    def getTableName(self):
        return "projects_projects_roles"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","projects_projects_roles","id") ]
    def getOrderBy(self):
        return [ "order_position", "id" ]
    def _before_insert(self,dbmgr=None):
        DBEObject._before_insert(self,dbmgr)
        query="select max(order_position) as order_position from %s" % dbmgr.buildTableName(self)
        tmp = dbmgr.select(self.getTableName(),query)
        if len(tmp)==1:
            tmp_position = tmp[0].getValue('order_position')
            if tmp_position is None:
                tmp_position = 0
            else:
                tmp_position = int(tmp_position)
            self.setValue('order_position', tmp_position + 1 )
        else:
            self.setValue('order_position', 1 )
dbschema['projects_projects_roles'] = DBEProjectProjectRole

class DBEProjectProject(DBEntity):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEntity.__init__(self, tablename, names, values, attrs, keys )
        if len(DBEProjectProject.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBEProjectProject.__columns[k]=parentcols[k]
            DBEProjectProject.__columns['project_id']=["uuid","not null default ''"]
            DBEProjectProject.__columns['project2_id']=["uuid","not null default ''"]
            DBEProjectProject.__columns['projects_projects_role_id']=["uuid","not null default ''"]
        self._columns=DBEProjectProject.__columns
    def getTableName(self):
        return "projects_projects"
    def getKeys(self):
        return { 'project_id':'uuid', 'project2_id':'uuid', 'projects_projects_role_id':'uuid' }
    def getFK(self):
        return [ ForeignKey("project_id","projects","id"), ForeignKey("project2_id","projects","id"), ForeignKey("projects_projects_role_id","projects_projects_roles","id") ]
    def getOrderBy(self):
        return [ "project_id", "project2_id", "projects_projects_role_id" ]
dbschema['projects_projects'] = DBEProjectProject

class DBETimetrack(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBETimetrack.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBETimetrack.__columns[k]=parentcols[k]
            DBETimetrack.__columns['fk_obj_id']=["uuid","default null"]
            DBETimetrack.__columns['fk_progetto']=["uuid","default null"]
            DBETimetrack.__columns['dalle_ore']=["datetime","not null default '0000-00-00 00:00:00'"]
            DBETimetrack.__columns['alle_ore']=["datetime","not null default '0000-00-00 00:00:00'"]
            DBETimetrack.__columns['ore_intervento']=["datetime","not null default '0000-00-00 00:00:00'"]
            DBETimetrack.__columns['ore_viaggio']=["datetime","not null default '0000-00-00 00:00:00'"]
            DBETimetrack.__columns['km_viaggio']=["int","not null default 0"]
            DBETimetrack.__columns['luogo_di_intervento']=["int","not null default 0"]
            DBETimetrack.__columns['stato']=["int","not null default 0"]
            DBETimetrack.__columns['costo_per_ora']=["float","not null default 0"]
            DBETimetrack.__columns['costo_valuta']=["varchar(255)","default null"]
        self._columns=DBETimetrack.__columns
    def getTableName(self):
        return "timetracks"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","timetracks","id"), ForeignKey("fk_obj_id","projects","id"), ForeignKey("fk_obj_id","folders","id"), ForeignKey("fk_obj_id","todo","id"), ForeignKey("fk_progetto","projects","id") ]
    def getOrderBy(self):
        return [ "fk_progetto", "stato desc", "dalle_ore desc", "fk_obj_id", "name" ]
dbschema['timetracks'] = DBETimetrack

class DBETodo(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBETodo.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBETodo.__columns[k]=parentcols[k]
            DBETodo.__columns['priority']=["int","not null default 0"]
            DBETodo.__columns['data_segnalazione']=["datetime","not null default '0000-00-00 00:00:00'"]
            DBETodo.__columns['fk_segnalato_da']=["uuid","default null"]
            DBETodo.__columns['fk_cliente']=["uuid","default null"]
            DBETodo.__columns['fk_progetto']=["uuid","default null"]
            DBETodo.__columns['fk_funzionalita']=["uuid","default null"]
            DBETodo.__columns['fk_tipo']=["uuid","default null"]
            DBETodo.__columns['stato']=["int","not null default 0"]
            DBETodo.__columns['descrizione']=["text","not null"]
            DBETodo.__columns['intervento']=["text","not null"]
            DBETodo.__columns['data_chiusura']=["datetime","not null default '0000-00-00 00:00:00'"]
        self._columns=DBETodo.__columns
    def getTableName(self):
        return "todo"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","todo","id"), ForeignKey("fk_segnalato_da","people","id"), ForeignKey("fk_cliente","companies","id"), ForeignKey("fk_progetto","projects","id"), ForeignKey("father_id","folders","id"), ForeignKey("father_id","todo","id"), ForeignKey("fk_tipo","todo_tipo","id") ]
    def getOrderBy(self):
        return [ "priority desc", "data_segnalazione desc", "fk_progetto", "name" ]
    def setDefaultValues(self,dbmgr):
        DBEObject.setDefaultValues(self,dbmgr)
        data_segnalazione = self.getValue('data_segnalazione')
        if data_segnalazione is None or data_segnalazione=="" or data_segnalazione=="00:00"\
                or self._formatDateTime(data_segnalazione).find("0000-00-00 00:00")>=0:
            self.setValue('data_segnalazione', self._getTodayString())
    def _before_insert(self,dbmgr=None):
        DBEObject._before_insert(self,dbmgr)
        data_segnalazione = self.getValue('data_segnalazione')
        if data_segnalazione is None or data_segnalazione=="" or data_segnalazione=="00:00"\
                or self._formatDateTime(data_segnalazione).find("0000-00-00 00:00")>=0:
            self.setValue('data_segnalazione', self._getTodayString())
        if dbmgr.getConnectionProvider().getDBType()=="POSTGRESQL":
            if self.getValue('data_segnalazione') in ['0000-00-00 00:00', '0000-00-00 00:00:00']:
                self.setValue('data_segnalazione',None)
            if self.getValue('data_chiusura') in ['0000-00-00 00:00', '0000-00-00 00:00:00']:
                self.setValue('data_chiusura',None)
        self._RULE_check_closed(dbmgr)
    def _before_update(self,dbmgr=None):
        DBEObject._before_update(self,dbmgr)
        if dbmgr.getConnectionProvider().getDBType()=="POSTGRESQL":
            if self.getValue('data_segnalazione') in ['0000-00-00 00:00', '0000-00-00 00:00:00']:
                self.setValue('data_segnalazione',None)
            if self.getValue('data_chiusura') in ['0000-00-00 00:00', '0000-00-00 00:00:00']:
                self.setValue('data_chiusura',None)
        self._RULE_check_closed(dbmgr)
        self._RULE_check_reopening(dbmgr)
    def _RULE_check_closed(self,dbmgr):
        data_chiusura = self.getValue('data_chiusura')
        if data_chiusura is None or data_chiusura==""\
                or self._formatDateTime(data_chiusura).find("0000-00-00 00:00")>=0:
            stato = 0
            try:
                stato = int( self.getValue('stato') )
            except Exception,e:
                pass
            if stato>=100:
                self.setValue('data_chiusura', dbmgr.getTodayString())
                self.setValue('stato', 100)
    def _RULE_check_reopening(self,dbmgr):
        data_chiusura = self.getValue('data_chiusura')
        if data_chiusura is None or data_chiusura==""\
                or self._formatDateTime(data_chiusura).find("0000-00-00 00:00")>=0:
            pass
        else:
            stato = 0
            try:
                stato = int( self.getValue('stato') )
            except Exception,e:
                pass
            if stato<100:
                self.setValue('data_chiusura', None)
dbschema['todo'] = DBETodo

class DBETodoTipo(DBEObject):
    __columns={}
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'} ):
        DBEObject.__init__(self, tablename, names, values, attrs, keys )
        if len(DBETodoTipo.__columns.keys())==0:
            parentcols=self.getColumns()
            for k in parentcols.keys():
                DBETodoTipo.__columns[k]=parentcols[k]
            DBETodoTipo.__columns['order_position']=["int","default 0"]
        self._columns=DBETodoTipo.__columns
    def getTableName(self):
        return "todo_tipo"
    def getKeys(self):
        return { 'id':'uuid' }
    def getFK(self):
        return [ ForeignKey("owner","users","id"), ForeignKey("group_id","groups","id"), ForeignKey("creator","users","id"), ForeignKey("last_modify","users","id"), ForeignKey("deleted_by","users","id"), ForeignKey("father_id","todo_tipo","id") ]
    def getOrderBy(self):
        return [ "order_position", "id" ]
    def getDefaultEntries(self):
        return [\
            {'order_position':1, 'name':'Analysis','owner':'-1','group_id':'-2','creator':'-1','last_modify':'-1','permissions':'rwxr--r--',},\
            {'order_position':2, 'name':'Design','owner':'-1','group_id':'-2','creator':'-1','last_modify':'-1','permissions':'rwxr--r--',},\
            {'order_position':3, 'name':'Development','owner':'-1','group_id':'-2','creator':'-1','last_modify':'-1','permissions':'rwxr--r--',},\
            {'order_position':4, 'name':'Test','owner':'-1','group_id':'-2','creator':'-1','last_modify':'-1','permissions':'rwxr--r--',},\
            {'order_position':5, 'name':'Bug','owner':'-1','group_id':'-2','creator':'-1','last_modify':'-1','permissions':'rwxr--r--',},\
            {'order_position':6, 'name':'Documentation','owner':'-1','group_id':'-2','creator':'-1','last_modify':'-1','permissions':'rwxr--r--',},\
            {'order_position':7, 'name':'Install','owner':'-1','group_id':'-2','creator':'-1','last_modify':'-1','permissions':'rwxr--r--',},\
        ]
dbschema['todo_tipo'] = DBETodoTipo

# #### Projects: end.


dbeFactory = DBEFactory(False)
for k in dbschema.keys():
    v = dbschema[k]
    dbeFactory.register( k, v )



# -*- coding: utf-8 -*-

#
# @copyright &copy; 2012 by Roberto Rocco Angeloni <roberto@roccoangeloni.it>
# @license http://opensource.org/licenses/lgpl-3.0.html GNU Lesser General Public License, version 3.0 (LGPLv3)
# @version $Id: __init__.py $
# @package dblayer
# 
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

from collections import OrderedDict
import datetime, os, string, sys, traceback, uuid

class DBLayerException(Exception):
    """Base Exception for the layer"""
    pass

class ForeignKey:
    """Maps a table's foreign key"""
    def __init__(self, colonna_fk="", tabella_riferita="", colonna_riferita=""):
        self.colonna_fk=colonna_fk
        self.tabella_riferita=tabella_riferita
        self.colonna_riferita=colonna_riferita
    def __str__(self):
        return "FK( %s => %s.%s )" % ( self.colonna_fk, self.tabella_riferita, self.colonna_riferita )

class DBEntity:
    """DB Entity
    Superclass for DB persistent entities"""
    def __init__(self, tablename=None, names=None, values=None, attrs=None, keys={'id':'number'},schema=None ):
        """Parameters:
            names=field names array
            values=field values array
            attrs=dict names->values
            keys=table's key field(s)
        """
        if attrs is not None:
            for k in attrs.keys():
                setattr( self , k, attrs[k] )
        elif names and values:
            self.__dict__ = {}
            for i in range(0,len(names)):
                if isDateTime(values[i]):
                    setattr( self, names[i] , string2datetime(values[ i ]) )
                else:
                    setattr( self, names[i] , values[ i ] )
        self._tablename = tablename
        self._keys = keys
        self._columns = {}
        self._schema = schema
        self._indexes = []
    def __str__(self):
        nomi = self.getNames()
        myattr={}
        for n in nomi:
            v = getattr(self, n)
            if v: myattr[n]=v
        ret = "%s( %s )" % ( self.getTypeName(), myattr )
        return ret
    def _formatDateTime(self, datetime):
        tmp = "%s" % datetime
        tmp = tmp.replace('/','-')
        return tmp[0:16]
    def _getTodayString(self):
        today = datetime.datetime.now()
        ret = "%s-%02d-%02d %02d:%02d:%02d" % (
            today.year,today.month,today.day,
            today.hour,today.minute,today.second,
        )
        return ret
    def getColumns(self):
        return self._columns
    def getDefaultEntries(self):
        """Returns a list of dictionaries, each dictionary defines an entry on the table"""
        return []
    def getTypeName(self):
        """Returns printable name of the type"""
        return self.__class__.__name__
    def getSchemaName(self):
        return self._schema
    def getTableName(self):
        return self._tablename
    def getOrderBy(self):
        return []
    def getOrderByString(self):
        return ",".join(self.getOrderBy())
    def getKeys(self):
        """Returns the keys"""
        return self._keys
    def getFK(self):
        """Returns an array with the foreign keys of the table"""
        return []
    def getFKForTable(self, tablename):
        """IF exists, returns an array with the FKs for the given table"""
        fks=self.getFK()
        ret=[]
        for f in fks:
            if f.tabella_riferita==tablename:
                ret.append(f)
        return ret
    def getFKDefinition(self,fk_column_name):
        """Returns all the FKs for the given table' column"""
        fks=self.getFK()
        ret=[]
        for f in fks:
            if f.colonna_fk==fk_column_name:
                ret.append(f)
        return ret
    def readFKFrom(self, dbe):
        """Reads the content of the referenced columns in the referenced table
        mapped in the given dbe"""
        fks = self.getFKForTable( dbe.getTableName() )
        for f in fks:
            v = dbe.getValue( f.colonna_riferita )
            if v:
                setattr( self, f.colonna_fk, v )
    def writeFKTo(self, masterdbe):
        """Writes the content of the referenced columns in the referenced table
        mapped in the given dbe"""
        fks = self.getFKForTable( masterdbe.getTableName() )
        for f in fks:
            v = self.getValue( f.colonna_fk )
            if v:
                masterdbe.setValue( f.colonna_riferita, v )
        return masterdbe
    def isPrimaryKey(self,field_name):
        if field_name in self.getKeys().keys():
            return True
        return False
    def isFK(self, field_name, tabella_riferita=None):
        """Tells if the given field is fk towards the given table"""
        fks = self.getFK()
        tmp = [ f.colonna_fk for f in fks if tabella_riferita is None or tabella_riferita==f.tabella_riferita ]
        return field_name in tmp
    def getNames(self):
        """Per entita' specializzate (ie mapping di una tabella) fare l'overload
        di questo metodo facendogli ritornare i nomi delle colonne."""
        ret=[]
        for d in self.__dict__.keys():
            if not d.startswith('_') and not d.startswith('get'): ret.append( d )
        return ret
    def getValue( self, chiave ):
        if not chiave.startswith("_") and chiave in self.__dict__:
            return getattr( self, chiave )
        else:
            return None
    def setValue( self, chiave, valore ):
        try:
            self.__dict__[chiave]=valore
        except Exception,e:
            if not chiave.startswith("_"):
                setattr( self, chiave, valore )
    def getValuesDictionary(self):
        """Returns a name:value dict"""
        ret = {}
        mynames = self.getNames()
        for m in mynames:
            v = self.getValue(m)
            if v:
                ret[m]=v
        return ret
    def setValuesDictionary(self,mydict):
        for k in mydict.keys():
            self.setValue(k, mydict[k])
    def cleanKeyFields( self ):
        """Resets all the key fields DBE
        Utility method for the <b>copy</b> method of DBMgr"""
        chiavi = self.getKeys()
        for c in chiavi:
            if self.__dict__.has_key( c ):
                self.__dict__.pop( c, None )
    def isNew(self):
        """True <=> all key fields are empty"""
        keys = self.getKeys()
        for c in keys:
            if not self.getValue(c) is None:
                if type(self.getValue(c))==str and self.getValue(c)>u'':
                    return False
                elif type(self.getValue(c))==unicode and self.getValue(c)>'':
                    return False
                elif type(self.getValue(c))==int or type(self.getValue(c))==float\
                        and self.getValue(c)!=0:
                    return False
                elif type(self.getValue(c))==bool:
                    return False
        return True
    def hasSameKeysOf(self, another_dbe):
        """Checks whether the keys of the given dbe and the current dbe have same values:
        no matter of tables"""
        ret = True
        keys = self.getKeys()
        for c in keys:
            if self.getValue(c) <> another_dbe.getValue(c):
                ret = False
                break
        return ret
    def uuid2hex(self,s):
        if s is None or len(s)==0:
            return s
        str_len = len(s)
        if str_len<4: return s
        if s[:4]=='uuid': return s
        ret = "uuid"
        for i in range(len(s)):
            ret += hex( ord( s[i] ) ).replace('0x','')
        return ret
    uuid2hex = classmethod(uuid2hex)
    def hex2uuid(self,a_str):
        if a_str[:4]!="uuid": return a_str
        s=a_str[4:]
        str_len = len(s)
        ret = ""
        for i in range(str_len/2):
            ret += chr(eval( "0x%s" % s[i*2:(i+1)*2] ))
        return ret
    hex2uuid = classmethod(hex2uuid)
    def validate(self):
        """Validate the DBE and applies custom logics"""
        pass
    def _before_insert(self,dbmgr=None):
        """Action to be taken before insert"""
        pass
    def _after_insert(self,dbmgr=None):
        """Action to be taken after insert"""
        pass
    def _before_update(self,dbmgr=None):
        """Action to be taken before update"""
        pass
    def _after_update(self,dbmgr=None):
        """Action to be taken after update"""
        pass
    def _before_delete(self,dbmgr=None):
        """Action to be taken before delete"""
        pass
    def _after_delete(self,dbmgr=None):
        """Action to be taken after delete"""
        pass
    def _before_copy(self,dbmgr=None):
        """Action to be taken before copy"""
        pass
    def _after_copy(self,dbmgr=None):
        """Action to be taken after copy"""
        pass

class DBAssociation(DBEntity):
    """Models N-M associations"""
    def __init__(self,tablename=None,names=None,values=None,attrs=None,keys=None,columns=[]):
        DBEntity.__init__(self,tablename,names,values,attrs,keys,columns)
    def getFromTableName(self):
        return self._fk[0].tabella_riferita
    def getToTableName(self):
        return self._fk[1].tabella_riferita

class DBConnectionProvider:
    """Superclass for all the connection providers.
    Subclasses of this may also implement connection pooling,
    not only specific DB connection"""
    def __init__(self, host, db, user, pwd, verbose=0):
        self._host = host
        self._db = db
        self._user = user
        self._pwd = pwd
        self._verbose = verbose
        self._conn = None
        self.customInit()
    def customInit(self):
        """Redefine this for idiosynchratic behaviour"""
        pass
    def getHost(self):
        return self._host
    def getUser(self):
        return self._user
    def getDB(self):
        return self._db
    def getConnection(self):
        return self._conn
    def freeConnection(self, conn):
        pass
    def getDBType(self):
        return "generic"
    def isProxy(self):
        return False
    def dbeType2dbType(self,dbetype):
        """TO OVERRIDE"""
        ret = dbetype
        if dbetype=="uuid" or dbetype==u"uuid":
            ret = "varchar(16)"
        return ret
    def dbType2dbeType(self,dbetype):
        """TO OVERRIDE"""
        ret = dbetype
        if dbetype=="varchar(16)":
            ret = "uuid"
        return ret
    def dbeConstraints2dbConstraints(self,constraints):
        return constraints
    def dbConstraints2dbeConstraints(t,definition):
        """ FIXME should this be in the connection object?
        Translates MySQL constraints into dbeContraints"""
        constraints = []
        if definition['Null']=='NO':
            constraints.append("not null")
        if not definition['Default'] is None:
            apice="'"
            if definition['Type'] in ['int(11)','float']:
                apice=""
            constraints.append( "default %s%s%s" % (apice,definition['Default'],apice) )
        elif definition['Default'] is None and definition['Null']=='YES':
            constraints.append( "default null" )
        return " ".join( constraints )
    dbConstraints2dbeConstraints = classmethod(dbConstraints2dbeConstraints)
    # TODO translate all those PHP functions: start.
    #static function dbColumnDefinition2dbeColumnDefinition($def) {
        #$constraints = DBEntity::dbConstraints2dbeConstraints($def);
        #return "'$col'=>array('".DBEntity::dbType2dbeType($def['Type'])."',"
                #.($constraints>'' ? $constraints : '')
                #."),\n";
    #}
    # TODO translate all those PHP functions: end.
    def getColumnsForTable(self,  tablename):
        """@return dictionary with column definitions, None if the table does not exists"""
        raise DBLayerException("DBConnectionProvider.getColumnsForTable: not implemented.")
    def getLocalFilePath(self):
        """@ret the root file path for file storage"""
        raise DBLayerException("DBConnectionProvider.getLocalFilePath: not implemented.")
    def uploadFile(self,local_filename):
        raise DBLayerException("DBConnectionProvider.uploadFile: not implemented.")
    def downloadFile(self,remote_filename,local_filename):
        """@par remote_filename
        @par local_filename where to download the file
        @ret local_filename if ok, None otherwise"""
        raise DBLayerException("DBConnectionProvider.downloadFile: not implemented.")

class DBMgr:
    def __init__(self, connProvider, verbose=0, schema = ''):
        self._connProvider = connProvider
        self._verbose = verbose
        self.conn = None
        self.lastMessages = None
        self._schema = schema
        self.user = None
        self.user_groups_list = []
        self.dbeFactory = None
        if self._connProvider.isProxy():
            self._connProvider.dbmgr = self

    def _description2names(self, desc):
        """Converts the resultset description (from the db api) in a names array"""
        return [d[0] for d in desc]
    def connect(self):
        """Open connection to the db"""
        self.conn = self._connProvider.getConnection()
    def disconnect(self):
        """Closes connection to the db"""
        self._connProvider.freeConnection( self.conn )
        self.conn = None
    def getConnectionProvider(self):
        return self._connProvider
    def getConnection(self):
        return self.conn
    def setDBEFactory( self, myfactory ):
        self.dbeFactory = myfactory
    def getDBEFactory(self):
        return self.dbeFactory
    def getRegisteredTypes(self):
        """Returns the registered types"""
        ret = {}
        tmp = self.getDBEFactory().getRegisteredTypes()
        for k in tmp.keys():
            if k!='default':
                ret[ k ] = tmp[k]
        if self._connProvider.isProxy():
            if len(ret.keys())>1 or ( len(ret.keys())==1 and ret.keys()[0]!='default' ):
                return ret
            ret =  self._connProvider.getRegisteredTypes(self)
        return ret
    def getClazz(self, typename):
        """Returns the registered type"""
        ret = self.getDBEFactory().getClazz(typename)
        if self._connProvider.isProxy() and ret==DBEntity:
            return self._connProvider.getClazz(self,typename)
        return ret
    def getClazzByTypeName(self, typename):
        """Returns the registered type"""
        ret = self.getDBEFactory().getClazzByTypeName(typename)
        if self._connProvider.isProxy() and ret==DBEntity:
            ret = self._connProvider.getClazzByTypeName(typename)
        return ret
    def initDB(self):
        """Initialize DB"""
        mytypes = self.dbeFactory.getRegisteredTypes()
        for tablename in mytypes.keys():
            if self._verbose: print "DBMgr.initDB: tablename = %s" % ( tablename )
            if tablename=='default':
                continue
            mydbe = mytypes[tablename]()
            # 1. Table definition
            self._createTable(mydbe)
            # 2. Default entries
            defaultEntries = mydbe.getDefaultEntries()
            if len(defaultEntries)==0: continue
            for d in defaultEntries:
                if self._verbose: print "DBMgr.initDB: d = %s" % (d)
                newdbe = mytypes[tablename]( attrs=d )
                try:
                    nuova = DBMgr.insert(self, newdbe)
                    print "DBMgr.initDB: nuova=%s (%s)" % (nuova,  newdbe)
                except Exception, e:
                    print "DBMgr.initDB: Exception = %s" % (e)
                    print "".join(traceback.format_tb(sys.exc_info()[2]))
    def _createTable(self, dbe):
        """Create the table for the given dbe"""
        myquery = self.dbe2sql(dbe)
        if self._verbose: print "DBMgr._createTable: myquery = %s" % ( myquery )
        try:
            cursor = self.getConnection().cursor()
            cursor.execute( myquery )
            numres = cursor.rowcount
            if self._verbose: print "DBMgr._createTable: numres = %s" % numres
            if self._connProvider.getDBType()=="POSTGRESQL":
                cursor.execute("commit")
        except Exception,e:
            if self._verbose:
                print "DBMgr._createTable: Exception = %s" % (e)
                print "".join(traceback.format_tb(sys.exc_info()[2]))
            if self._connProvider.getDBType()=="POSTGRESQL":
                cursor.execute( "rollback" )
    def getColumnsForTable(self, tablename):
        return self.getConnectionProvider().getColumnsForTable(tablename)
    def checkDB(self):
        """Checks whether the db and the dbe defs in the dbe factory are sync'd"""
        mytypes = self.dbeFactory.getRegisteredTypes()
        for tablename in mytypes.keys():
            if tablename=='default':
                continue
            mydbe = mytypes[tablename]()
            # 1. Check table definition TODO
            mytablename = self.buildTableName(mydbe)
            columns = self.getColumnsForTable( mytablename )
            if columns is None:
                # Table does not exists
                if self._verbose: print "DBMgr.checkDB: Table '%s' does not exists." % (mytablename)
                myquery = self.dbe2sql(mydbe)
                try:
                    cursor = self.getConnection().cursor()
                    cursor.execute( myquery )
                    numres = cursor.rowcount
                    print "DBMgr.checkDB: numres = %s" % numres
                    if self._connProvider.getDBType()=="POSTGRESQL":
                        cursor.execute("commit")
                except Exception,e:
                    if self._verbose:
                        print "DBMgr.checkDB: Exception = %s" % (e)
                        print "".join(traceback.format_tb(sys.exc_info()[2]))
            else:
                #print "DBMgr.checkDB: columns=%s" % ( [x for x in columns.keys()] )
                columnsdefinitions = mydbe.getColumns()
                # Columns to add
                delta = [x for x in columnsdefinitions.keys() if not x in columns.keys()]
                for c in delta:
                    myquery = "alter table %s add column %s %s" % ( self._buildTableName(mydbe), c, self.column2sql( columnsdefinitions[c], mydbe.isPrimaryKey(c) ) )
                    try:
                        cursor = self.getConnection().cursor()
                        cursor.execute( myquery )
                        if self._connProvider.getDBType()=="POSTGRESQL":
                            cursor.execute("commit")
                    except Exception,e:
                        if self._verbose or True:
                            print "DBMgr.checkDB: Exception = %s" % (e)
                            print "".join(traceback.format_tb(sys.exc_info()[2]))
                # Columns to drop
                delta = [ x for x in columns.keys() if not x in columnsdefinitions.keys() ]
                for c in delta:
                    myquery = "alter table %s drop column %s %s" % ( self._buildTableName(mydbe), c, self.column2sql( columnsdefinitions[c], mydbe.isPrimaryKey(c) ) )
                    try:
                        cursor = self.getConnection().cursor()
                        cursor.execute( myquery )
                        if self._connProvider.getDBType()=="POSTGRESQL":
                            cursor.execute("commit")
                    except Exception,e:
                        if self._verbose:
                            print "DBMgr.checkDB: Exception = %s" % (e)
                            print "".join(traceback.format_tb(sys.exc_info()[2]))
            # 2. Check default entries
            defaultEntries = mydbe.getDefaultEntries()
            if len(defaultEntries)==0: continue
            for d in defaultEntries:
                if self._verbose: print "DBMgr.checkDB: d = %s" % (d)
                cerca = mytypes[tablename]( attrs=d )
                tmp = DBMgr.search(self, cerca, uselike=False)
                if len(tmp)==0:
                    try:
                        nuova = DBMgr.insert(self, cerca)
                    except Exception,e:
                        if self._verbose:
                            print "DBMgr.checkDB: Exception = %s" % (e)
                            print "".join(traceback.format_tb(sys.exc_info()[2]))
    def removeDB(self):
        """BEWARE: Remove all the tables and all the rows from the DB!!!"""
        mytypes = self.dbeFactory.getRegisteredTypes()
        for tablename in mytypes.keys():
            if tablename=='default':
                continue
            mydbe = mytypes[tablename]()
            myquery = "drop table %s" % ( self._buildTableName(mydbe) )
            if self._verbose: print "DBMgr.removeDB: myquery = %s" % ( myquery )
            try:
                cursor = self.getConnection().cursor()
                cursor.execute( myquery )
                numres = cursor.rowcount
                if self._verbose: print "DBMgr.removeDB: numres = %s" % numres
            except Exception,e:
                if self._verbose: print "DBMgr.removeDB: Exception = %s" % (e)
                if self._connProvider.getDBType()=="POSTGRESQL":
                    cursor.execute( "rollback" )
    def dbe2sql(self,dbe):
        """Creates the 'create table' statement"""
        ret = []
        cols = dbe.getColumns()
        ret.append( "--" )
        ret.append( "-- %s" % ( dbe.getTypeName() ) )
        ret.append( "--" )
        ret.append( "create table %s (" % ( self._buildTableName(dbe) ) )
        for c in cols.keys():
            ret.append( " %s %s," % ( c, self.column2sql( cols[c], dbe.isPrimaryKey(c) ) ) )
        ret.append( " primary key(%s)" % ( ",".join( dbe.getKeys().keys() ) ) )
        ret.append( ");" )
        ret.append("")
        return "\n".join(ret)
    def column2sql(self, mydef,  isPrimaryKey):
            mytype = self._connProvider.dbeType2dbType( mydef[0] )
            constraints = ""
            if len(mydef)>1:
                constraints = mydef[1]
            not_null=""
            if isPrimaryKey:
                not_null="not null"
            if len(constraints)==0:
                constraints = not_null
            constraints = self._connProvider.dbeConstraints2dbConstraints(constraints)
            return "%s %s" % (mytype,constraints)
    def _escapeString(self,s):
        return "'%s'" % s.replace("\\","\\\\").replace("\'","''")
    def _formatDateTime(self, datetime):
        tmp = "%s" % datetime
        tmp = tmp.replace('/','-')
        # 2013-04-09: start.
        if len(tmp)==8:
            return "0000-00-00 %s" % (tmp)
        # 2013-04-09: end.
        return tmp[0:16]
    def getTodayString(self):
        today = datetime.datetime.now()
        ret = "%s-%02d-%02d %02d:%02d:%02d" % (
            today.year,today.month,today.day,
            today.hour,today.minute,today.second,
        )
        return ret
    def buildTableName(self,dbe):
        return self._buildTableName(dbe)
    def _buildTableName(self,dbe):
        tablename=''
        __schema = dbe.getSchemaName()
        if __schema is None or __schema=="":
            __schema = self._schema
        if __schema>"":
            tablename = "%s_%s" % ( __schema, dbe.getTableName() )
        return tablename
    def _buildKeysCondition(self, dbe):
        ret = []
        chiavi = dbe.getKeys()
        for c in chiavi.keys():
            if chiavi[c]=='number':
                ret.append( "%s=%s" % ( c, getattr(dbe,c) ) )
            else:
                ret.append( "%s='%s'" % ( c, getattr(dbe,c) ) )
            pass
        return string.join(ret, ' AND ')
    def buildKeysCondition(self, dbe):
        return self._buildKeysCondition(dbe)
    def _buildInsertString(self,dbe):
        nomicampi = dbe.getNames()
        tmpnomi=[]
        tmpvalori=[]
        for n in nomicampi:
            v = dbe.getValue(n)
            if v is None: continue
            tmpnomi.append(n)
            if isinstance(v,str) or isinstance( v, unicode ):
                if not isDateTime(v):
                    tmpvalori.append( "%s" % ( self._escapeString( v ) ) )
                else:
                    tmpvalori.append( "'%s'" % ( v ) )
            elif v.__class__.__name__=='DateTime' or isinstance(v,datetime.datetime):
                tmpvalori.append( "'%s'" % ( self._formatDateTime(v) ) )
            elif isinstance(v,bool):
                if v:
                    tmpvalori.append('1')
                else:
                    tmpvalori.append('0')
            else:
                    tmpvalori.append( "%s" % (v) )
        query = "insert into %s (%s) values (%s)" % ( self._buildTableName(dbe), string.join(tmpnomi,","), string.join(tmpvalori,",") )
        return query
    def buildInsertString(self,dbe):
        return self._buildInsertString(dbe)
    def _buildUpdateString(self,dbe):
        nomicampi = dbe.getNames()
        chiavi = dbe.getKeys()
        setstring=[]
        for n in nomicampi:
            v = dbe.getValue(n)
            if not (n in chiavi):
                if v is None:
                    continue
                elif isinstance(v,str) or isinstance( v, unicode ):
                    if not isDateTime(v):
                        setstring.append( "%s=%s" % ( n, self._escapeString( v ) ) )
                    else:
                        setstring.append( "%s='%s'" % ( n, v ) )
                elif isinstance(v,bool):
                    if v:
                        setstring.append( "%s=1" % (n) )
                    else:
                        setstring.append( "%s=0" % (n) )
                elif v.__class__.__name__=='DateTime' or isinstance(v,datetime.datetime):
                    setstring.append( "%s='%s'" % ( n, self._formatDateTime(v) ) )
                elif v.__class__.__name__=='Time' or isinstance(v,datetime.time):
                    setstring.append( "%s='%s'" % ( n, self._formatDateTime(v) ) )
                else:
                    setstring.append( "%s=%s" % ( n, v ) )
                # 2012.04.04: start.
                # 2012.04.04: not set in the same method of PHP
                #else:
                    #setstring.append( "%s=NULL" % (n) )
                # 2012.04.04: end.
        query = "update %s set %s where %s" % ( self._buildTableName(dbe),\
                                                 string.join(setstring,", "),\
                                                 self._buildKeysCondition(dbe) )
        return query
    def _buildWhereCondition( self, dbe, uselike=1,case_sensitive=True):
        nomicampi = dbe.getNames()
        clausole=[]
        for n in nomicampi:
            is_from = n[0:5]=='from_'
            is_to = n[0:3]=='to_'
            v = dbe.getValue(n)
            if v is None:
                if self._connProvider.getDBType()=="POSTGRESQL":
                    clausole.append( "%s is null" % (n) )
            elif v=='null':
                clausole.append( "%s is null" % (n) )
            elif isinstance( v, str ) or isinstance( v, unicode ):
                if is_from:
                    if len(v)>0: clausole.append( "%s>='%s'" % (n[5:],v) )
                elif is_to:
                    if len(v)>0: clausole.append( "%s<='%s'" % (n[3:],v) )
                elif v.find('0000-00-00 00:00')>=0:
                    clausole.append( "(%s='%s' or %s is null)" % (n,v,n) )
                elif uselike:
                    if len(v)>0:
                        myv = "%%%s%%" % v
                        if v.find('*')>=0:
                            myv = "%s" % ( v.replace("*","%%") )
                        if case_sensitive:
                            clausole.append( "%s like '%s'" % (n,myv) )
                        else:
                            clausole.append( "lower(%s) like '%s'" % (n,myv.lower()) )
                else:
                    clausole.append( "%s='%s'" % (n,v) )
            else:
                if is_from:
                    if not v is None: clausole.append( "%s>=%s" % (n[5:],v) )
                elif is_to:
                    if not v is None: clausole.append( "%s<=%s" % (n[3:],v) )
                else:
                    clausole.append( "%s=%s" % (n,v) )
        return clausole
    def _buildSelectString(self, dbe, uselike=1,case_sensitive=True):
        clausole = self._buildWhereCondition( dbe, uselike, case_sensitive )
        if len(clausole)>0:
            query = "select * from %s where %s" % ( self._buildTableName(dbe), string.join(clausole," and ") )
        else:
            query = "select * from %s" % ( self._buildTableName(dbe) )
        return query
    def _buildDeleteString(self,dbe):
        query = "delete from %s where %s" % ( self._buildTableName(dbe), self._buildKeysCondition(dbe) )
        return query
    def _before_insert(self, dbe):
        """Action to be taken before insert"""
        return dbe
    def _after_insert(self, dbe):
        """Action to be taken after insert"""
        return dbe
    def _before_update(self, dbe):
        """Action to be taken before update"""
        return dbe
    def _after_update(self, dbe):
        """Action to be taken after update"""
        return dbe
    def _before_delete(self, dbe):
        """Action to be taken before delete"""
        return dbe
    def _after_delete(self, dbe):
        """Action to be taken after delete"""
        return dbe
    def _before_copy(self, dbe):
        """Action to be taken before copy"""
        return dbe
    def _after_copy(self, dbe):
        """Action to be taken after copy"""
        return dbe
    def insert(self, dbe, cleanKeysIfError=True):
        """Insert the new dbe into the db"""
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self._connProvider.insert(dbe)
        # Before Insert
        dbe._before_insert(self)
        dbe = self._before_insert(dbe)
        # Insert
        query = self._buildInsertString(dbe)
        if self._verbose: print "DBMgr.insert: query=%s" % ( query )
        try:
            cursor = self.getConnection().cursor()
            cursor.execute( query )
            self.getConnection().commit()
            # After Insert
            dbe = self._after_insert(dbe)
            dbe._after_insert(self)
        except Exception, e:
            if self._verbose:
                print "DBMgr.insert: Probblemi: %s" % (e)
                print "DBMgr.insert: self._connProvider = %s - %s" % ( self._connProvider._db, self._connProvider )
                print "DBMgr.insert: query=%s" % (query)
                print "".join(traceback.format_tb(sys.exc_info()[2]))
            try:
                ris = self.getConnection().execute( query )
                self.getConnection().commit()
                # After Insert
                dbe = self._after_insert(dbe)
                dbe._after_insert(self)
            except Exception, e1:
                if cleanKeysIfError: dbe.cleanKeyFields()
                if self._verbose:
                    print dir(self.getConnection())
                    print "DBMgr.insert: Probblemi: %s" % (e1)
                    print "DBMgr.insert: self._connProvider = %s - %s" % ( self._connProvider._db, self._connProvider )
                    print "DBMgr.insert: query=%s" % (query)
                    print "".join(traceback.format_tb(sys.exc_info()[2]))
                raise DBLayerException("DBMgr.insert: unable to execute - %s.Error cause: %s" % ( query, e ))
        return dbe
    def update(self,dbe):
        """Update the given dbe"""
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self._connProvider.update(dbe)
        # Before Update
        dbe._before_update(self)
        dbe = self._before_update(dbe)
        # Update
        query = self._buildUpdateString(dbe)
        if self._verbose: print "DBMgr.update: query=%s" % ( query )
        try:
            cursor = self.getConnection().cursor()
            cursor.execute( query )
            self.getConnection().commit()
            # After Update
            dbe = self._after_update(dbe)
            dbe._after_update(self)
        except Exception, e:
            if self._verbose:
                print "DBMgr.update: Probblemi: %s" % (e1)
                print "DBMgr.update: query=%s" % (query)
                print "".join(traceback.format_tb(sys.exc_info()[2]))
            try:
                ris = self.getConnection().execute( query )
                self.getConnection().commit()
                # After Update
                dbe = self._after_update(dbe)
                dbe._after_update(self)
            except Exception, e1:
                dbe = None
                if self._verbose:
                    print "DBMgr.update: Probblemi: %s" % (e1)
                    #print "DBMgr.update: self._connProvider = %s - %s" % ( self._connProvider._db, self._connProvider )
                    print "DBMgr.update: query=%s" % (query)
                    print "".join(traceback.format_tb(sys.exc_info()[2]))
                raise DBLayerException("DBMgr.update: unable to execute\n\t%s.Error cause: %s" % ( query, e1 ))
        return dbe
    def select(self,tablename,searchString):
        """Returns a list of DBE according to the searchString"""
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self._connProvider.select(tablename,searchString)
        cursor = self.getConnection().cursor()
        ret = []
        try:
            cursor.execute( searchString )
            numres = cursor.rowcount
            names = None
            try:
                names = self._description2names(cursor.description)
                if self._connProvider.getDBType()=="MYSQL":
                    if self._verbose: print "DBMgr.select: found %s rows." % ( numres )
                    for i in range( 0, numres ):
                        valori = cursor.fetchone()
                        ret.append( self.getDBEFactory()( tablename, names=names, values=valori ) )
                else:
                    listavalori = cursor.fetchall()
                    numres = len(listavalori)
                    if self._verbose: print "DBMgr.select: found %s rows." % ( numres )
                    for i in range( numres ):
                        ret.append( self.getDBEFactory()( tablename, names=names, values=listavalori[i] ) )
            except TypeError,e:
                if self._verbose:
                    print "DBMgr.select: ECCEZIONE=%s (%s)" % (e,searchString)
                    print "".join(traceback.format_tb(sys.exc_info()[2]))
        except Exception,e:
            if self._connProvider.getDBType()=="POSTGRESQL":
                cursor.execute( "rollback" )
            if self._verbose:
                print "DBMgr.select: ECCEZIONE=%s (%s)" % (e,searchString)
                print "".join(traceback.format_tb(sys.exc_info()[2]))
            raise e
        cursor.close()
        return ret
    def delete(self,dbe):
        """Delete the given dbe"""
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self._connProvider.delete(dbe)
        # Before Delete
        dbe._before_delete(self)
        dbe = self._before_delete(dbe)
        # Delete
        query = "delete from %s where %s" % ( self._buildTableName(dbe), self._buildKeysCondition(dbe) )
        if self._verbose: print "Delete: query=%s" % query
        try:
            ris = self.getConnection().execute( query )
            self.getConnection().commit()
            # After Delete
            dbe = self._after_delete(dbe)
            dbe._after_delete(self)
        except Exception, e:
            if self._verbose:
                print "DBMgr.delete: ECCEZIONE=%s (%s)" % (e,searchString)
                print "".join(traceback.format_tb(sys.exc_info()[2]))
            try:
                ris = self.getConnection().query( query )
                self.getConnection().commit()
                # After Delete
                dbe = self._after_delete(dbe)
                dbe._after_delete(self)
            except Exception, e:
                dbe = None
                raise DBLayerException("DBMgr.delete: unable to execute\n\t%s.Error cause: %s" % ( query, e ))
        return dbe
    def search(self, dbe, uselike=1, orderby=None, ignore_deleted=True, full_object=True):
        """Provides Search methods"""
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self._connProvider.search(dbe, uselike, orderby, ignore_deleted, full_object)
        query = self._buildSelectString(dbe,uselike)
        if orderby is not None: query += " ORDER BY %s" % (orderby)
        if self._verbose: print "DBMgr.search: query=%s" % query
        return self.select( dbe.getTableName(), query )
    def getNextId(self, dbe):
        nomeTabella = "%s_seq_id" % ( self._schema )
        tmp = self.select(nomeTabella, "select id from %s where name=''" % ( nomeTabella ) )
        myid = 1
        if len(tmp)>0:
            myid = tmp[0].getValue('id') + 1
            self.select(nomeTabella, "insert into %s (id,name) values (%s,'')" % (nomeTabella,myid) )
        else:
            self.select(nomeTabella, "update %s set id=%s where name=''" % (nomeTabella,myid) )
        return myid
    def getNextUuid(self, dbe, length=16):
        return ( ("%s" % uuid.uuid4()).replace('-','') )[:16]
    def copy(self, dbe):
        """Copy the given DBE"""
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self._connProvider.copy(dbe)
        # 1. Read the whole DBE from db
        tmp = self.search( dbe, uselike=0 )
        if len(tmp)!=1: raise DBLayerException("Unable to copy the given DBE; found %s instances." % ( len(tmp) ) )
        mydbe = tmp[0]
        # 2. Resetting keys
        mydbe.cleanKeyFields()
        dbe = mydbe
        # 3. Insert a new entry
        dbe._before_copy( self )
        dbe = self._before_copy( dbe )
        dbe = self.insert( dbe )
        dbe = self._after_copy( dbe )
        dbe._after_copy( self )
        return dbe
    def searchByKeys(self,dbe):
        """Returns a list of dbe using the keys of the dbe in input as a search key
        Typically will return a list with zero or one element"""
        chiavi = dbe.getKeys()
        # 2012.06.04: start.
        # 2012.06.04: this works better with NoSQL dbs
        cerca = self.getClazzByTypeName(dbe.getTypeName())()
        for k in chiavi.keys():
            cerca.setValue(k,  dbe.getValue(k))
        cerca.setValuesDictionary(chiavi)
        ret = self.search(cerca, uselike=0)
        #query = "select * from %s where %s" % (self._buildTableName(dbe), self._buildKeysCondition(dbe))
        #ret = self.select( dbe.getTableName(), query )
        # 2012.06.04: end.
        return ret
    def exists(self,dbe):
        """Tells if another record exists in the database with the same key values
        as the dbe in input"""
        if dbe.isNew():
            return False
        tmp = self.searchByKeys(dbe)
        return len(tmp)>0
    def isDateTime(self,s):
        return isDateTime(s)
    def datetime2string(self, dt):
        tmpdata = "%s" % ( dt )
        tmpdata = "%s-%s-%s 00:00" % (tmpdata[:4],tmpdata[5:7],tmpdata[8:10])
        return tmpdata.replace('/','-')
    def string2datetime(self, dt):
        return DateTime(dt)
    def millis2datetime(self, millis):
        return DateTime(millis)
    def convertDate(self, dt):
        tmpdata = "%s" % ( dt )
        tmpdata = "%s-%s-%s" % (tmpdata[8:10],tmpdata[5:7],tmpdata[:4])
        return tmpdata.replace('/','-')
    def getLastMessages(self):
        return self.lastMessages
    def ping(self):
        if self._connProvider.isProxy():
            return self._connProvider.ping()
        return "pong"
    def _loadUserGroups(self):
        if self.user is None:
            return
        cerca = self.getClazzByTypeName('DBEUserGroup')()
        cerca.readFKFrom(self.user)
        lista=self.search( cerca, uselike=0 )
        lista_gruppi=[]
        for g in lista:
            lista_gruppi.append( "%s"%g.getValue('group_id') )
        if not "%s"%self.user.getValue('group_id') in lista_gruppi:
            lista_gruppi.append( "%s"%self.user.getValue('group_id') )
        self.setUserGroupsList(lista_gruppi)
    def login(self,user,pwd):
        if self._connProvider.isProxy():
            self.user = self._connProvider.login(user,pwd)
            if self.user is None:
                return None
            self._loadUserGroups()
            return self.user
        if pwd is None or len(pwd)==0 or user is None or len(user)==0:
            raise DBLayerException("Missing username or password")
        cerca = self.getClazzByTypeName('DBEUser')(attrs={'login':user,'pwd':pwd})
        ret = []
        try:
            ret = self.search(cerca,uselike=False)
        except Exception,e:
            try:
                self.initDB()
                newuser = self.getClazzByTypeName('DBEUser')(attrs={'login':user,'pwd':pwd})
                newuser = self.insert(newuser)
                searchGroup = self.getClazzByTypeName('DBEGroup')(attrs={'name':user})
                newgroup = self.search(searchGroup,uselike=False)[0]
                if self._verbose: print "DBMgr.login: newuser=%s" % ( newuser )
                newfolder = self.getClazzByTypeName('DBEFolder')(attrs={\
                    'owner':newuser.getValue('id'),\
                    'group_id':newgroup.getValue('id'),\
                    'creator':newuser.getValue('id'),\
                    'last_modify':newuser.getValue('id'),\
                    'name':user})
                newfolder = self.insert(newfolder)
                if self._verbose: print "DBMgr.login: newfolder=%s" % ( newfolder )
                ret = self.search(cerca,uselike=False)
            except Exception,e1:
                if self._verbose: print "DBMgr.login: ECCEZIONE=%s" % ( e )
                if self._verbose: print "".join(traceback.format_tb(sys.exc_info()[2]))
        if len(ret)==1:
            self.user = ret[0]
            self._loadUserGroups()
        return self.user
    def relogin(self):
        myuser = self.getDBEUser()
        if myuser is None:
            return False
        return self.login( myuser.getValue('login'), myuser.getValue('pwd') )
    def getServerIDString(self):
        """Returns a string uniquely identifying the user and the connection"""
        d = self.getConnectionProvider().getDBType()
        u = "nobody"
        if not self.getDBEUser() is None:
            u = self.getDBEUser().getValue('login')
        h = "%s" % self.getConnectionProvider().getHost()
        if d=="SQLite":
            h = ( "%s" % self.getConnectionProvider().getDB() ).split( os.path.sep )[-1]
        if h.find("://")>0:
            _url = h.split("://")
            if len(_url)>1:
                h = "%s:%s" % ( _url[0],_url[1].split("/")[0] )
        return "%s@%s" % (u,h)
    def getDBEUser(self):
        return self.user
    def setDBEUser(self,dbeuser):
        self.user = dbeuser
    def isLoggedIn(self):
        remoteuser = self.getDBEUser()
        if self._connProvider.isProxy():
            remoteuser = self._connProvider.getLoggedUser()
        return not remoteuser is None
    def getUserGroupsList(self):
        return self.user_groups_list
    def setUserGroupsList(self,user_groups_list):
        self.user_groups_list = user_groups_list
    def hasGroup(self,group_id):
        """ The logged user has the requested group_id ? """
        return "%s"%group_id in self.user_groups_list
    def addGroup(self,group_id):
        if not group_id in self.user_groups_list:
            self.user_groups_list.append(group_id)
    def getFormSchema(self,language='python',aClassname=''):
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self._connProvider.getFormSchema(language,aClassname)
        raise DBLayerException("DBMgr.getFormSchema: TODO")
    def getDBSchema(self,language='python',aClassname=''):
        if self._connProvider.isProxy():
            # 2012.04.02: start.
            if not self.isLoggedIn() and not self.relogin():
                raise DBLayerException("Proxy connection down!")
            # 2012.04.02: end.
            return self._connProvider.getDBSchema(language,aClassname)
        raise DBLayerException("DBMgr.getDBSchema: TODO")
    def uploadFile(self,local_filename):
        return self._connProvider.uploadFile(local_filename)
    def downloadFile(self,remote_filename,local_filename,view_thumbnail=False):
        return self._connProvider.downloadFile(remote_filename,local_filename,view_thumbnail)

class DBEFactory:
    """Returns the correct class for the given tablename"""
    def __init__(self, verbose = False):
        self._cache = OrderedDict() #{}
        self._cache_by_typename = OrderedDict() #{}
        self.register('default', DBEntity)
        self.verbose=verbose
    def register(self, tablename, clazz):
        self._cache[ tablename ]=clazz
        self._cache_by_typename[ clazz().getTypeName() ]=clazz
    def getRegisteredTypes(self):
        return self._cache
    def getClazz(self, tablename):
        ret = None
        if self._cache.has_key( tablename ):
            ret = self._cache[tablename]
        else:
            ret=self._cache['default']
        return ret
    def getClazzByTypeName(self, typename, case_sensitive=True):
        if self.verbose: print "DBEFactory.getClazzByTypeName: typename=%s" % ( typename )
        ret = self._cache['default']
        if case_sensitive:
            if self._cache_by_typename.has_key(typename):
                ret = self._cache_by_typename[typename]
        else:
            chiavi = [ k.lower() for k in self._cache_by_typename.keys() ]
            typename_lower = typename.lower()
            if typename_lower in chiavi:
                for k in self._cache_by_typename.keys():
                    if k.lower()==typename_lower:
                        ret = self._cache_by_typename[k]
                        break
        if self.verbose: print "DBEFactory.getClazzByTypeName: ret=%s" % ( ret )
        return ret
    def __call__(self, tablename, names=None, values=None, *args):
        """Some kind of magic..."""
        if self.verbose:
            print "DBEFactory.__call__: tablename=%s" % ( tablename )
        myclazz = self.getClazz(tablename)
        ret = myclazz( tablename, names, values )
        if self.verbose:
            print "DBEFactory.__call__: Classe: %s\tIstanza: %s" % ( myclazz, ret )
        return ret

def isDateTime(s):
    # 2005-04-12 00:00
    # 2012/1/21 17:51:56
    try:
        if s[4]=='-' and s[7]=='-' and ( s[13]==':' or len(s)==10 ):
            return True
        else:
            data,ora = s.split(" ")
            if ( len( data.split("/") )==3 or len( data.split("-") )==3
                    ) and (
                    len(ora.split(":"))>=2 or len(ora.split("."))>=2
                    ):
                return True
            return False
    except Exception,e:
        return False
def string2datetime(s):
    if len(s)>=15: #len(s)>=16:
        data,ora = s.split(" ")
        _y,_m,_d = 0,0,0
        if data.find("/")>=0:
            _y,_m,_d = data.split("/")
        elif data.find("-")>=0:
            _y,_m,_d = data.split("-")
        if ora.find(":")>=0:
            _h,_M = ( ora.split(":") )[:2]
        elif ora.find(".")>=0:
            _h,_M = ( ora.split(".") )[:2]
        try:
            _y,_m,_d,_h,_M = int(_y),int(_m),int(_d),int(_h),int(_M)
        except Exception, e:
            #print "s:", s
            return None
        if _y==0 and _m==0 and _d==0 and _h==0 and _M==0:
            return None
        if _y==0 and _m==0 and _d==0:
            return datetime.time(_h,_M)
        try:
            return datetime.datetime(_y,_m,_d,_h,_M)
        except Exception,e:
            raise DBLayerException("string2datetime(%s): exception=%s" % (s,e))
    else:
        if s.find("/")>=0:
            _y,_m,_d = s.split("/")
        elif s.find("-")>=0:
            _y,_m,_d = s.split("-")
        _y,_m,_d = int(_y),int(_m),int(_d)
        if _y==0 and _m==0 and _d==0:
            return None
        return datetime.datetime(_y,_m,_d)

def createConnection(aUrl,aVerbose=False):
    myconn = None
    myurl = aUrl
    if myurl.startswith("http"):
        if myurl.endswith("/xmlrpc_server.php"):
            pass # OK
        elif myurl.endswith("/"):
            myurl = "%sxmlrpc_server.php" % ( myurl )
        else:
            myurl = "%s/xmlrpc_server.php" % ( myurl )
        from rprj.dblayer.xmlrpc import XmlrpcConnectionProvider
        myconn = XmlrpcConnectionProvider(myurl,'','','',aVerbose) #1)
    elif myurl.startswith("json"):
        # Quick patch: if there is a port, then don't add the default page
        if not myurl.find(":"):
            if myurl.endswith("/jsonserver.php"):
                pass # OK
            elif myurl.endswith("/"):
                myurl = "%sjsonserver.php" % ( myurl )
            else:
                myurl = "%s/jsonserver.php" % ( myurl )
        from rprj.dblayer.jsonconn import JsonConnectionProvider
        myconn = JsonConnectionProvider(myurl,'','','',aVerbose) #1)
    elif myurl.startswith("mysql"):
        xx,host, db, user, pwd = myurl.split(":")
        from rprj.dblayer.mydb import MYConnectionProvider
        myconn = MYConnectionProvider(host, db, user, pwd,aVerbose)
    elif myurl.startswith("postgresql"):
        xx,host, db, user, pwd = myurl.split(":")
        from rprj.dblayer.pgdb import PGConnectionProvider
        myconn = PGConnectionProvider(host, db, user, pwd,aVerbose)
    elif myurl.startswith("sqlite"):
        tmp_path=myurl.split(":")
        path = "".join(tmp_path[1:])
        if len(tmp_path)>2:
            tmp_path[1] = "%s:"%tmp_path[1]
            path = "".join(tmp_path[1:])
        if path.startswith("///"):
            path = path[2:]
        from rprj.dblayer.sqlitedb import SQLiteConnectionProvider
        myconn = SQLiteConnectionProvider('',path,'','',aVerbose)
    return myconn




"""
http://docs.python.org/distutils/setupscript.html
http://docs.python.org/distutils/builtdist.html

c:\Python27\python.exe setup.py bdist_wininst --install-script=rPrj_postinstall.py

# Update PyPi
python setup.py register
python setup.py sdist upload
# Update installations
sudo easy_install --upgrade r-prj
"""

import sys
from distutils.core import setup

my_scripts = ['rPrj.pyw',]
my_data_files = []
if sys.platform=='win32':
	my_data_files.append( ('Icons',['rprj/formulator/icons/rprj.ico']) )
	my_scripts.append('rPrj_postinstall.py')
elif sys.platform=='darwin':
	#my_data_files.append( ('Resources/RPrj.app/Contents',['RPrj.app/Contents/Info.plist']) )
	#my_data_files.append( ('Resources/RPrj.app/Contents/MacOS',['RPrj.app/Contents/MacOS/osx.sh']) )
	#my_data_files.append( ('Resources/RPrj.app/Contents/Resources',['RPrj.app/Contents/Resources/logo_32x32.icns']) )
	pass
else:
	my_data_files.append( ('/usr/share/icons',['rprj/formulator/icons/rprj.png']) )
	my_data_files.append( ('/usr/share/applications',['rprj.desktop']) )

setup(name='R-Prj'
	,version="0.5.6"
	,description='R-Prj framework libraries and client application'
	,license="http://opensource.org/licenses/lgpl-3.0.html GNU Lesser General Public License, version 3.0 (LGPLv3)"
	,author='Roberto Rocco Angeloni'
	,author_email='roberto@roccoangeloni.it'
	,url='http://www.kware.it/main.php?obj_id=uuid69742e34663137643339336236653937'
#	,url='https://sourceforge.net/projects/r-prj/'
#	,download_url='https://sourceforge.net/projects/r-prj/'
	,requires=['pyqt',]
	,packages=['rprj'
		,'rprj/apps'
		,'rprj/dblayer'
		,'rprj/formulator'
		,'rprj/net'
		,'rprj/qtwidgets'
	]
	,package_data={'rprj/qtwidgets': ['rprj/qtwidgets/*.png'], 'rprj/apps/images': ['rprj/apps/images/linneighborhood.png']}
	,data_files=my_data_files
#		('/usr/share/icons',['rprj/apps/icons/rprj.png'])
#		,('/usr/share/applications',['rprj.desktop'])
#			('bitmaps',['rprj/qtwidgets/format-fill-color.png'
#				,'rprj/qtwidgets/format-indent-less.png'
#				,'rprj/qtwidgets/format-indent-more.png'
#				,'rprj/qtwidgets/'
#				]
#			),
#	]
	,scripts=my_scripts
)



# -*- coding: utf-8 -*-
#
# @copyright &copy; 2011 by Roberto Rocco Angeloni <roberto@roccoangeloni.it>
# @license http://opensource.org/licenses/lgpl-3.0.html GNU Lesser General Public License, version 3.0 (LGPLv3)
# @version $Id: sqlitedb.py $
# @package dblayer
# 
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import os,sys

from rprj.dblayer import *

major_v,minor_v, subminor_v,etichetta,altri = sys.version_info
if major_v==2 and minor_v==4:
	from pysqlite2 import dbapi2 as sqlite
else:
	from sqlite3 import dbapi2 as sqlite

def convert_text(s):
	# XXX do not use Unicode
	return s.decode("utf-8")
sqlite.register_converter("TEXT", convert_text)


class SQLiteConnectionProvider(DBConnectionProvider):
	"""SQLite Connection provider"""
	def __init__(self, host, db, user, pwd, verbose=0):
		"""@param host?
		@param db: path al file
		@param user?
		@param pwd?
		"""
		DBConnectionProvider.__init__(self,  host, db, user, pwd, verbose )
	def customInit(self):
		"""Redefine this for idiosynchratic behaviour"""
		pass
	def getConnection(self):
		if self._conn is None:
			try:
				self._conn = sqlite.connect(self._db)#, detect_types=sqlite.PARSE_DECLTYPES )
				# 2012.03.16: start.
				# 2012.03.16: FIXED - Now handles non utf8 encoded strings :-)))
				self._conn.text_factory = lambda x: unicode(x, "utf-8", "ignore")\
						.replace(u"\\r",u"\r")\
						.replace(u"\\n",u"\n")\
						.replace(u"\\t",u"\t")\
						.replace(u"\\\"",u"\"")\
						.replace(u"\\\\",u"\\")
				# 2012.03.16: end.
			except Exception, e:
				if self._verbose: print "SQLiteConnectionProvider.getConnection: db=%s Error=%s" % (self._db,e)
		return self._conn
	def freeConnection(self,conn):
		try:
			conn.close()
		except Exception, e:
			if self._verbose: print "SQLiteConnectionProvider.freeConnection: Error=%s" % (e)
		del conn
		del self._conn
		self._conn = None
	def getDBType(self):
		return "SQLite"
	def getColumnsForTable(self,  tablename):
		"""@return dictionary with column definitions, None if the table does not exists"""
		if self._verbose:
			#print "SQLiteConnectionProvider.getColumnsForTable: tablename=%s" % (tablename)
			#print "SQLiteConnectionProvider.getColumnsForTable: %s" % dir(self._conn)
			pass
		cursor = self.getConnection().cursor()
		ret = {}
		searchString = "select * from %s where 1=0" % tablename
		try:
			cursor.execute( searchString )
			for d in cursor.description:
				ret[ d[0] ] = [ d[0], d[1], d[2], d[3], d[4], d[5], d[6] ]
				#if self._verbose:
				#	print "SQLiteConnectionProvider.getColumnsForTable: d=(%s,%s,%s,%s,%s,%s,%s) %s" % (d[0], d[1], d[2], d[3], d[4], d[5], d[6], len(d))
		except sqlite.OperationalError, oe:
				# Table does not exists
				#print "SQLiteConnectionProvider.getColumnsForTable: OperationalError=%s (%s)" % (oe,searchString)
				ret = None
		except Exception,e:
			if self._verbose:
				print "SQLiteConnectionProvider.getColumnsForTable: ECCEZIONE=%s (%s)" % (e,searchString)
				print "".join(traceback.format_tb(sys.exc_info()[2]))
			raise e
		finally:
			cursor.close()
		#print "SQLiteConnectionProvider.getColumnsForTable: ret=%s" % (ret)
		return ret
	def getLocalFilePath(self):
		dest_dir = "%s%sfiles" % ( os.path.sep.join( self.getDB().split( os.path.sep )[:-1] ), os.path.sep)
		return dest_dir
	def uploadFile(self,local_filename):
		dest_dir = "%s%sfiles" % ( os.path.sep.join( self.getDB().split( os.path.sep )[:-1] ), os.path.sep)
		if not os.path.exists(dest_dir): os.mkdir(dest_dir)
		if self._verbose: print "SQLiteConnectionProvider.uploadFile: dest_dir:",dest_dir
		# Short local filename
		filename_corto = local_filename[ len(os.path.dirname(local_filename)): ]
		if filename_corto.startswith('/'):
			filename_corto = filename_corto[1:]
		if self._verbose: print "SQLiteConnectionProvider.uploadFile: filename_corto:",filename_corto
		dest_file = "%s%s%s" % (dest_dir,os.path.sep,filename_corto)
		if self._verbose: print "SQLiteConnectionProvider.uploadFile: dest_file:",dest_file
		# Load file
		myfile = file( local_filename, 'rb' ).read()
		# Upload
		file(dest_file,'wb').write( myfile )
		return dest_file
	def downloadFile(self,remote_filename,local_filename,view_thumbnail=False):
		"""@par uuid
		@par local_filename where to download the file
		@ret local_filename if ok, None otherwise"""
		file(local_filename,'wb').write( file(remote_filename,'rb').read() )
		return local_filename
